from __future__ import annotations

import logging
from copy import copy

import numpy as np
from scipy import sparse

from skretrieval.retrieval import ForwardModel, Minimizer, RetrievalTarget


class Rodgers(Minimizer):
    def __init__(
        self,
        max_iter: int = 10,
        lm_damping: float = 0,
        iterative_update_lm: bool = False,
        retreat_lm: bool = False,
        lm_change_factor: float = 1.5,
        convergence_factor: float = 1,
        convergence_check_method="linear",
        lm_damping_method="fletcher",
        apply_cholesky_scaling=False,
    ):
        """
        Implements the standard inverse problem method described in "Inverse Methods for Atmospheric Sounding"
        by Rodgers (2000).

        Parameters
        ----------
        max_iter: int, optional
            The maximum number of iterations to perform when calling retrieve. Default: 10
        lm_damping: float, optional
            The Levenberg-Marquardt damping parameter.  A value of 0 would indicate no damping.  Default: 0
        iterative_update_lm: bool, optional
            If True, the LM damping factor is modified each iteration based on how well the problem is converging.
            Default: False
        retreat_lm: bool, optional
            If True, if the chi sq is worse for one iteration the state vector is retreated back to the previous
            iteration's value. Default: False
        lm_change_factor: float, optional
            The multiplicative factor applied to the LM damping parameter each iteration if iterative_update_lm is
            set to True.  Default: 1.5
        convergence_factor: float, optional
            See convergence_check_method.  Reasonable values are 1.01 if convergence_check_method is linear, and 1
            if convergence_check_method is dcost.  Default: 1
        convergence_check_method: str, optional
            Sets the method of checking for convergence.  If 'linear' then (expected chi_sq) / (chi_sq) is checked if
            it is less than convergence_factor. If 'dcost', then the derivative of the cost function is checked against
            convergence_factor.
        lm_damping_method: str, optional
            One of 'fletcher', 'prior', or 'identity'.  If 'fletcher', the LM damping term will be
            lm_damping * diag(K^T inv_Sy K).  If 'prior', the lm damping is lm_damping * inv_Sa.
            If 'identity' the damping is lm_damping * identity. Default is 'fletcher'
        apply_cholesky_scaling: bool, optional
            If True, then attempts will be made to rescale the state vector and measurement vectors to a more
            numerically suitable space
        """
        self._max_iter = max_iter
        self._lm_damping = lm_damping
        self._iterative_update_lm = iterative_update_lm
        self._retreat_lm = retreat_lm
        self._lm_change_factor = lm_change_factor
        self._convergence_factor = convergence_factor
        self._convergence_check_method = convergence_check_method
        self._lm_damping_method = lm_damping_method
        self._apply_cholesky_scaling = apply_cholesky_scaling

    @staticmethod
    def _measurement_parameters(retrieval_target: RetrievalTarget, measurement_l1):
        """
        Calculates parameters necessary for the iteration that depend on the measurement data.

        Parameters
        ----------
        retrieval_target: RetrievalTarget
            Target that is being retrieved
        measurement_l1:
            Measurement l1 data

        Returns
        -------
        y_meas_dict: dict
            Dictionary containing various parameters of the measurement vector
        y_meas: np.array
            The measurement vector, length (m,)
        Sy: matrix like
            Matrix of size (m,m) that contains the error covariance of the measurements
        inv_Sy: matrix like
            Matrix of size (m,m), that is the inverse of inv_Sy
        good_meas: np.array
            Flag of which measurements are used in y_meas.  y_meas = y_meas_dict['y'][good_meas]
        """
        y_meas_dict = retrieval_target.measurement_vector(measurement_l1)

        y_meas = y_meas_dict["y"]
        good_meas = ~np.isnan(y_meas)
        y_meas = y_meas[good_meas]

        if "y_error" in y_meas_dict:
            # Have measurement error
            if len(np.shape(y_meas_dict["y_error"])) == 1:
                # Only supplied is the diagonal of the error elements
                Sy = sparse.csc_matrix(
                    sparse.diags(y_meas_dict["y_error"][good_meas], 0)
                )
                inv_Sy = sparse.csc_matrix(
                    sparse.diags(1 / y_meas_dict["y_error"][good_meas], 0)
                )
            else:
                Sy = y_meas_dict["y_error"][np.ix_(good_meas, good_meas)]
                inv_Sy = np.linalg.inv(Sy)
        else:
            # No user supplied error, use identity matrix
            Sy = sparse.csc_matrix(sparse.eye(len(y_meas), len(y_meas)))
            inv_Sy = sparse.csc_matrix(sparse.eye(len(y_meas), len(y_meas)))

        return y_meas_dict, y_meas, Sy, inv_Sy, good_meas

    @staticmethod
    def _apriori_parameters(retrieval_target: RetrievalTarget):
        """
        Calculates several parameters related to the apriori using the retrieval target.

        Parameters
        ----------
        retrieval_target: RetrievalTarget

        Returns
        -------
        x_a: np.array
            Array of length (n,) that is the apriori state
        inv_Sa: matrix_like
            Matrix of size (n,n) that is the apriori covariance
        initial_guess: np.array
            Array of length (n,) that is the current state
        """
        x_a = retrieval_target.apriori_state()
        initial_guess = retrieval_target.state_vector()

        inv_Sa = retrieval_target.inverse_apriori_covariance()

        if inv_Sa is None:
            # No apriori covariance/regularization
            # Use initial guess to make the matrices as x_a might be None as well
            inv_Sa = np.zeros((len(initial_guess), len(initial_guess)))

        if x_a is None:
            x_a = np.zeros_like(initial_guess)

        return x_a, inv_Sa, initial_guess

    def retrieve(
        self,
        measurement_l1,
        forward_model: ForwardModel,
        retrieval_target: RetrievalTarget,
    ):
        retrieval_target.initialize(forward_model, measurement_l1)

        output_dict = {}

        y_meas_dict, y_meas, Sy, inv_Sy, good_meas = self._measurement_parameters(
            retrieval_target, measurement_l1
        )
        x_a, inv_Sa, initial_guess = self._apriori_parameters(retrieval_target)

        xs = []
        xs.append(initial_guess)
        ys = []
        chi_sq_prev = None
        best_x = retrieval_target.state_vector()
        best_chi_sq = 1e99

        if self._apply_cholesky_scaling:
            x_scaler_inv = sparse.diags(np.sqrt(inv_Sa.diagonal()))
            y_scaler = sparse.diags(1 / np.sqrt(inv_Sy.diagonal()))
            y_scaler_inv = sparse.diags(np.sqrt(inv_Sy.diagonal()))
            x_scaler = sparse.diags(1 / np.sqrt(inv_Sa.diagonal()))
        else:
            x_scaler_inv = sparse.eye(len(x_a))
            x_scaler = x_scaler_inv
            y_scaler = sparse.eye(len(y_meas))
            y_scaler_inv = sparse.eye(len(y_meas))

        y_meas = y_scaler_inv @ y_meas

        inv_Sa = x_scaler @ inv_Sa @ x_scaler
        inv_Sy = y_scaler @ inv_Sy @ y_scaler
        x_a = x_scaler_inv @ x_a

        for iter_idx in range(self._max_iter):
            x = x_scaler_inv @ retrieval_target.state_vector()
            forward_l1 = forward_model.calculate_radiance()

            y_ret_dict = retrieval_target.measurement_vector(forward_l1)

            K = y_ret_dict["jacobian"]

            K = y_scaler_inv @ K[good_meas, :] @ x_scaler
            y_ret = y_scaler_inv @ y_ret_dict["y"][good_meas]

            ys.append(y_ret_dict["y"])

            approx_hessian = K.T @ inv_Sy @ K

            # Left side of rodgers equation
            if sparse.issparse(K):
                if self._lm_damping_method.lower() == "fletcher":
                    A = (
                        approx_hessian
                        + self._lm_damping * sparse.diags((approx_hessian).diagonal())
                        + inv_Sa
                    )
                elif self._lm_damping_method.lower() == "prior":
                    A = approx_hessian + (self._lm_damping + 1) * inv_Sa
                elif self._lm_damping_method == "identity":
                    A = (
                        approx_hessian
                        + inv_Sa
                        + self._lm_damping * np.eye(inv_Sa.shape)
                    )
                else:
                    msg = "lm_damping_method should be one of fletcher, prior, or identity"
                    raise ValueError(msg)
            else:
                if self._lm_damping_method.lower() == "fletcher":
                    A = (
                        approx_hessian
                        + self._lm_damping * np.diag(np.diag(approx_hessian))
                        + inv_Sa
                    )
                elif self._lm_damping_method.lower() == "prior":
                    A = approx_hessian + (self._lm_damping + 1) * inv_Sa
                elif self._lm_damping_method == "identity":
                    A = (
                        approx_hessian
                        + inv_Sa
                        + self._lm_damping * np.eye(inv_Sa.shape[0])
                    )
                else:
                    msg = "lm_damping_method should be one of fletcher, prior, or identity"
                    raise ValueError(msg)

            A_without_lm = approx_hessian + inv_Sa

            # Right side of rodgers equation
            B = K.T @ inv_Sy @ (y_meas - y_ret) - inv_Sa @ (x - x_a)

            if sparse.issparse(A):
                try:
                    delta_x = np.linalg.solve(A.toarray(), B)
                except np.linalg.LinAlgError:
                    delta_x = np.linalg.lstsq(A.toarray(), B)[0]
            else:
                try:
                    delta_x = np.linalg.solve(A, B)
                except np.linalg.LinAlgError:
                    delta_x = np.linalg.lstsq(A, B)[0]

            if sparse.issparse(A_without_lm):
                try:
                    delta_x_without_lm = np.linalg.solve(A_without_lm.toarray(), B)
                except np.linalg.LinAlgError:
                    delta_x_without_lm = np.linalg.lstsq(A_without_lm.toarray(), B)[0]
            else:
                try:
                    delta_x_without_lm = np.linalg.solve(A_without_lm, B)
                except np.linalg.LinAlgError:
                    delta_x_without_lm = np.linalg.lstsq(A_without_lm, B)[0]

            x_new = x_scaler @ (x + delta_x)

            chi_sq_only_meas = (y_meas - y_ret).T @ inv_Sy @ (y_meas - y_ret)
            chi_sq = chi_sq_only_meas + (x_a - x).T @ inv_Sa @ (x_a - x)

            chi_sq_only_meas_linear = (
                (y_meas - y_ret - K @ delta_x_without_lm).T
                @ inv_Sy
                @ (y_meas - y_ret - K @ delta_x_without_lm)
            )

            chi_sq_linear = (y_meas - y_ret - K @ delta_x_without_lm).T @ inv_Sy @ (
                y_meas - y_ret - K @ delta_x_without_lm
            ) + (x_a - x - delta_x_without_lm).T @ inv_Sa @ (
                x_a - x - delta_x_without_lm
            )

            chi_sq_only_meas /= len(y_meas)
            chi_sq /= len(y_meas)

            if np.isnan(chi_sq) or np.isinf(chi_sq):
                msg = "chi_sq is infinite or nan"
                raise ValueError(msg)

            chi_sq_only_meas_linear /= len(y_meas)
            chi_sq_linear /= len(y_meas)

            logging.info(
                "",
                extra={
                    "chi_sq": chi_sq,
                    "chi_sq_only_meas": chi_sq_only_meas,
                    "expected_chi_sq": chi_sq_linear,
                    "chi_sq_only_meas_linear": chi_sq_only_meas_linear,
                },
            )

            retrieval_target.update_state(x_new)

            dcost = (
                delta_x_without_lm
                @ (K.T @ inv_Sy @ (y_meas - y_ret) + inv_Sa @ (x_a - x))
                / len(x)
            )
            logging.info("", extra={"dcost": dcost})

            if chi_sq_prev is not None and chi_sq_prev < chi_sq:
                # Iteration was worse
                if self._iterative_update_lm:
                    self._lm_damping *= self._lm_change_factor**2
                if self._retreat_lm:
                    retrieval_target.update_state(x_scaler @ best_x)
                else:
                    chi_sq_prev = chi_sq
                logging.info(
                    "Iteration was worse increasing LM factor",
                    extra={"lm_damping": self._lm_damping},
                )
            elif chi_sq_prev is None or chi_sq < chi_sq_prev:
                best_x = copy(x)
                best_chi_sq = chi_sq
                chi_sq_prev = best_chi_sq
                if self._iterative_update_lm and iter_idx > 0:
                    self._lm_damping /= self._lm_change_factor
                    logging.info(
                        "Iteration was better decreasing LM factor",
                        extra={"lm_damping": self._lm_damping},
                    )

            xs.append(retrieval_target.state_vector())

            if self._convergence_check_method.lower() == "linear":
                if (
                    chi_sq / chi_sq_linear < self._convergence_factor
                    and chi_sq / chi_sq_linear > 1
                ):
                    logging.info(
                        "Stopping due to early convergence",
                        extra={"convergence_ratio": chi_sq / chi_sq_linear},
                    )
                    break
            elif (self._convergence_check_method.lower() == "dcost") and (
                dcost < self._convergence_factor
            ):
                logging.info(
                    "Stopping due to early convergence. dcost is less than convergence_factor"
                )
                break

            if iter_idx != self._max_iter - 1:
                predicted_delta_y = y_meas - y_ret - K @ delta_x_without_lm
                retrieval_target.adjust_parameters(
                    forward_model,
                    y_ret_dict,
                    chi_sq,
                    chi_sq_linear,
                    iter_idx,
                    predicted_delta_y,
                )

                if retrieval_target.state_vector_allowed_to_change():
                    x_a, inv_Sa, initial_guess = self._apriori_parameters(
                        retrieval_target
                    )

                    x_a = x_scaler_inv @ x_a

                    inv_Sa = x_scaler @ inv_Sa @ x_scaler
                if retrieval_target.measurement_vector_allowed_to_change():
                    (
                        y_meas_dict,
                        y_meas,
                        Sy,
                        inv_Sy,
                        good_meas,
                    ) = self._measurement_parameters(retrieval_target, measurement_l1)
                    y_meas = y_scaler_inv @ y_meas
                    inv_Sy = y_scaler @ inv_Sy @ y_scaler

        if self._max_iter > 0:
            # Calculate the solution covariance and averaging kernels
            try:
                if sparse.issparse(A_without_lm):
                    S = np.linalg.inv(A_without_lm.toarray())
                else:
                    S = np.linalg.inv(A_without_lm)
            except np.linalg.LinAlgError:
                if sparse.issparse(A_without_lm):
                    S = np.linalg.pinv(A_without_lm.toarray())
                else:
                    S = np.linalg.pinv(A_without_lm)

            G = S @ K.T @ inv_Sy
            A = G @ K
            meas_error_covar = G @ (Sy.dot(G.T))
        else:
            S = np.zeros((len(x_a), len(x_a)))
            G = None
            A = np.zeros((len(x_a), len(x_a)))
            meas_error_covar = np.zeros((len(x_a), len(x_a)))

        output_dict["xs"] = xs
        output_dict["gain_matrix"] = G
        output_dict["averaging_kernel"] = A

        output_dict["error_covariance_from_noise"] = meas_error_covar
        output_dict["solution_covariance"] = S

        output_dict["ys"] = ys
        output_dict["y_meas"] = y_meas_dict["y"]

        if self._max_iter > 0:
            output_dict["chi_sq_meas"] = chi_sq_only_meas
            output_dict["chi_sq_meas_linear"] = chi_sq_only_meas_linear

            output_dict["chi_sq"] = chi_sq
            output_dict["chi_sq_linear"] = chi_sq_linear
        else:
            output_dict["chi_sq_meas"] = None
            output_dict["chi_sq_meas_linear"] = None

            output_dict["chi_sq"] = None
            output_dict["chi_sq_linear"] = None

        return output_dict



from __future__ import annotations

import unittest

import numpy as np
import sasktran as sk

from skretrieval.core import OpticalGeometry
from skretrieval.core.lineshape import DeltaFunction, Gaussian
from skretrieval.legacy.core.sensor.spectrograph import Spectrograph


class TestSpectrographIntegration(unittest.TestCase):
    def setUp(self):
        self.optical_geometry = OpticalGeometry(
            observer=np.array([0, 6392 * 1000, 6392 * 1000]),
            look_vector=np.array([0, -1, 0]),
            local_up=np.array([0, 0, 1]),
            mjd=54372,
        )

        self.wavel_nm = np.arange(300, 330, 1)

        line_shape = Gaussian(fwhm=1, mode="linear")
        self.spectrograph = Spectrograph(
            self.wavel_nm, line_shape, DeltaFunction(), DeltaFunction()
        )

    def test_radiance(self):
        geo = sk.Geometry()
        geo.lines_of_sight = self.spectrograph.measurement_geometry(
            self.optical_geometry
        )

        atmo = sk.Atmosphere()
        atmo["air"] = sk.Species(sk.Rayleigh(), sk.MSIS90())
        atmo["ozone"] = sk.Species(sk.O3DBM(), sk.Labow())

        engine = sk.EngineHR(atmosphere=atmo, geometry=geo)
        engine.options["numordersofscatter"] = 1

        engine.wavelengths = self.spectrograph.required_wavelengths(0.1)

        rad = engine.calculate_radiance()

        _ = self.spectrograph.model_radiance(
            self.optical_geometry, engine.wavelengths, engine.geometry, rad
        )

    def test_weighting_function(self):
        geo = sk.Geometry()
        geo.lines_of_sight = self.spectrograph.measurement_geometry(
            self.optical_geometry
        )

        atmo = sk.Atmosphere()
        atmo["air"] = sk.Species(sk.Rayleigh(), sk.MSIS90())
        atmo["ozone"] = sk.Species(sk.O3DBM(), sk.Labow())

        atmo.wf_species = "ozone"

        engine = sk.EngineHR(atmosphere=atmo, geometry=geo)
        engine.options["numordersofscatter"] = 1

        engine.wavelengths = self.spectrograph.required_wavelengths(0.1)

        rad, wf = engine.calculate_radiance()

        _ = self.spectrograph.model_radiance(
            self.optical_geometry, engine.wavelengths, engine.geometry, rad, wf
        )



from __future__ import annotations

import unittest

import numpy as np
import sasktran as sk

from skretrieval.core.lineshape import Gaussian, Rectangle
from skretrieval.core.platform.satellite import fake_satellite_scan
from skretrieval.legacy.core.sensor.spectrograph import Spectrograph
from skretrieval.legacy.core.simulator import MeasurementSimulator


class TestMeasSimulator(unittest.TestCase):
    def test_basic(self):
        axis, sun = fake_satellite_scan(altitudes=[20, 25, 30, 35, 40])

        atmo = sk.Atmosphere()
        atmo["air"] = sk.Species(sk.Rayleigh(), sk.MSIS90())
        atmo["ozone"] = sk.Species(sk.O3DBM(), sk.Labow())

        sensor = Spectrograph(
            np.arange(300, 330, 1),
            pixel_shape=Gaussian(fwhm=1, mode="linear"),
            vert_fov=Gaussian(fwhm=0.001),
            horiz_fov=Rectangle(width=1, mode="constant"),
        )

        simulator = MeasurementSimulator(
            sensor, axis, atmo, options={"numordersofscatter": 1}
        )

        simulator.engine.geometry.sun = sun

        simulator.calculate_radiance()



from __future__ import annotations

import abc
from typing import Iterable

import numpy as np
import xarray as xr


class StateVectorElement(abc.ABC):
    """
    A state vector element is a component of the full state vector used in the retrieval. Each state vector element
    has a state, and a prior state/covariance associated with it.  The state vector element must also be able
    to update itself, calculate the jacobian matrix for itself.
    """

    def __init__(self, enabled: bool = True):
        self._enabled = enabled

    @abc.abstractmethod
    def state(self) -> np.array:
        pass

    def inverse_apriori_covariance(self) -> np.ndarray:
        n = len(self.state())
        return np.zeros((n, n))

    def apriori_state(self) -> np.array:
        return np.zeros_like(self.state())

    def lower_bound(self) -> np.array:
        n = len(self.state())
        return np.ones(n) * (-np.inf)

    def upper_bound(self) -> np.array:
        n = len(self.state())
        return np.ones(n) * (np.inf)

    @property
    def enabled(self) -> bool:
        return self._enabled

    @enabled.setter
    def enabled(self, e: bool):
        self._enabled = e

    @abc.abstractmethod
    def name(self) -> str:
        pass

    @abc.abstractmethod
    def propagate_wf(self, radiance: xr.Dataset) -> xr.Dataset:
        pass

    @abc.abstractmethod
    def update_state(self, x: np.array):
        pass

    def modify_input_radiance(self, radiance: xr.Dataset):
        return radiance


class StateVector:
    def __init__(self, elements: Iterable[StateVectorElement]):
        """
        A full state vector made up of a collection of state vector elements.

        Parameters
        ----------
        elements: Iterable[StateVectorElement]
            A collection of state vector elements
        """
        self._elements = elements

    @property
    def state_elements(self):
        return self._elements

    def update_sasktran_radiance(self, radiance: xr.Dataset, drop_old_wf: bool = False):
        """
        Modifies radiances output from sasktran based on the state vector elements if applicable, e.g., if a state
        vector element is a wavelength shift this will apply it.

        Propagates weighting functions from the sasktran radiance raw output to weighting functions for each
        state vector element.

        If drop_old_wf is set to true then the old weighting functions are removed from the radiance.

        Parameters
        ----------
        radiance: xr.Dataset
            Output from sk.Engine.calculate_radiance(output_format='xarray')
        drop_old_wf: bool, Optional
            If true then the old weighting functions are removed after being propagated to the state vector. Default
            False

        Returns
        -------
        radiance: xr.Dataset
            Modified radiance with a new key 'wf' that is the jacobian with respect to the full state vector.
        """
        all_jacobian = []
        for state_element in self._elements:
            all_jacobian.append(state_element.propagate_wf(radiance))
            radiance = state_element.modify_input_radiance(radiance)

        new_wf = xr.concat(all_jacobian, dim="x")
        radiance["wf"] = new_wf

        if drop_old_wf:
            wf_names = [key for key in radiance if key.startswith("wf_")]
            radiance = radiance.drop(wf_names)
        return radiance



from __future__ import annotations

from abc import ABC, abstractmethod

import numpy as np

from skretrieval.core.radianceformat import RadianceBase


class ForwardModel(ABC):
    """
    A ForwardModel is an object which is capable of calculating a radiance.  This serves as the primary interface
    to the retrieval, along with the RetrievalTarget.
    """

    @abstractmethod
    def calculate_radiance(self):
        pass


class RetrievalTarget(ABC):
    """
    The retrieval target defines the parameter that is to be retrieved, and also what measurements are going to be
    used to retrieve it.  Notation is similar to that of Rodgers.
    """

    @abstractmethod
    def state_vector(self):
        """
        Returns
        -------
        np.array
            The state vector, x
        """

    @abstractmethod
    def measurement_vector(self, l1_data: RadianceBase):
        """

        Parameters
        ----------
        l1_data: RadianceBase
            Radiance data.  Usually this is an instrument specific instance of RadianceBase, and the RetrievalTarget
            only works with specific formats.

        Returns
        -------
        dict
            Keys 'y' for the measurement vector, 'jacobian' for the jacobian of the measurement vector (if weighting
            functions are in l1_data, 'y_error' the covariance of 'y' (if error information is provided in l1_data)
        """

    @abstractmethod
    def update_state(self, x: np.ndarray):
        """
        Updates the state for the new state vector.  Note that this change has to propagate backwards to the ForwardModel
        somehow.  Typically this is done by passing a climatology into the RetrievalTarget at initiliazation which is
        used in the ForwardModel.

        Parameters
        ----------
        x: np.array
            New state vector
        """

    @abstractmethod
    def apriori_state(self) -> np.array:
        """
        Returns
        -------
        np.array
            Apriori state vector, x_a.  If no apriori is used return None
        """

    @abstractmethod
    def inverse_apriori_covariance(self):
        """
        Returns
        -------
        np.array
            Inverse of the apriori covariance matrix.  If no apriori is used return None.
        """

    def initialize(  # noqa: B027
        self, forward_model: ForwardModel, meas_l1: RadianceBase
    ):
        """
        Called at the beginning of the retrieval and can be used to initialize parameters

        Parameters
        ----------
        forward_model
        meas_l1

        """

    @staticmethod
    def state_vector_allowed_to_change():
        """
        Returns
        -------
        bool
            True if the state vector/apriori may change shape between iterations, False otherwise.
        """
        return False

    @staticmethod
    def measurement_vector_allowed_to_change():
        """
        Returns
        -------
        bool
            True if the measurement_vector may change shape between iterations, False otherwise.
        """
        return False

    def adjust_parameters(
        self,
        forward_model,  # noqa: ARG002
        y_dict,  # noqa: ARG002
        chi_sq,  # noqa: ARG002
        chi_sq_linear,  # noqa: ARG002
        iter_idx,  # noqa: ARG002
        predicted_delta_y,  # noqa: ARG002
    ):
        return None


class Minimizer(ABC):
    """
    A class which performs minimization between some aspect of measurement level1 data and the forward model simulations.
    """

    @abstractmethod
    def retrieve(
        self,
        measurement_l1: RadianceBase,
        forward_model: ForwardModel,
        retrieval_target: RetrievalTarget,
    ):
        """

        Parameters
        ----------
        measurement_l1: RadianceBase
            The data we are trying to match, either from a real instrument or simulations.
        forward_model: ForwardModel
            A model for the data in measurement_l1
        retrieval_target: RetrievalTarget
            What we are trying to retrieve

        Returns
        -------
        dict
            Various parameters specific to the minimizer
        """



from __future__ import annotations

import logging
import math

import numpy as np

from skretrieval.geodetic import geodetic

from .rotationmatrix import RotationMatrix, UnitVectors


# -----------------------------------------------------------------------------
#           Class PlatformPointing
# -----------------------------------------------------------------------------
class PlatformPointing:

    """
    A class used to manage the pointing requirements of an **instrument** mounted on a **platform**. The primary objective
    is to

    * Mount an instrument on a platform, eg balloon, aircraft, satellite, ground.
    * Place and orient the platform at a location in space, on the ground or in the atmosphere.
    * Generate unit vectors at the front aperture of the instrument representing lines of sight and similar that can be
      used by the radiative transfer codes.

    The PlatformPointing class generates rotation matrices to transform between the different standard control frames as well as
    methods to mount the instrument on the platform, position the platform and rotate the platform.
    """

    # -----------------------------------------------------------------------------
    #           __init__
    # -----------------------------------------------------------------------------

    def __init__(self):
        self._ICF_Init_PCF: RotationMatrix = RotationMatrix(
            array=[[1, 0, 0], [0, -1, 0], [0, 0, -1]]
        )  # Initial matrix to get ICF properly aligned with PCF. ICFx is +PCFx, ICFy is -PCFy and ICFz is -PCFz
        self._PCF_Init_GCF: RotationMatrix = RotationMatrix(
            array=[[0, 1, 0], [1, 0, 0], [0, 0, -1]]
        )  # Initial matrix to get PCF properly aligned with GCF. PCF X is GCFy (due North), PCF y is GCFx (due East), PCF Z is -GCFz (upwards)
        self._IRM: RotationMatrix = (
            RotationMatrix()
        )  # Instrument Rotation Matrix. Applies instrument internal mirrors and turntables tec,
        self._ICF_to_PCF: RotationMatrix = (
            RotationMatrix()
        )  # Rotation Matrix from Instrument Control Frame to Platform Control Frame. Applies monutnig information
        self._PCF_to_GCF: RotationMatrix = (
            RotationMatrix()
        )  # Rotation Matrix from Platform Control Frame to Geodetic Control Frame. E.G. Applies the Yaw, Pitch, Roll to orient the platform in local geodetic space.
        self._GCF_to_ECEF: RotationMatrix = (
            RotationMatrix()
        )  # Rotation Matrix to convert local geodetic coordinates to geographic geocentric. ECEF
        self._utc: np.datetime64 = (
            None  #: Platform pointing also stores the UTC time as a convenience
        )
        self._geo: geodetic = geodetic()  #: An instance of sasktran Geodetic object.
        self._geolocation: np.ndarray = np.zeros([3])
        self._local_west: np.ndarray = None
        self._local_south: np.ndarray = None
        self._local_up: np.ndarray = None
        self._nadirmode: bool = False

    # ------------------------------------------------------------------------------
    #           mount_instrument_on_platform
    # ------------------------------------------------------------------------------
    def mount_instrument_on_platform(
        self,
        azimuth_degrees: float,
        elevation_degrees: float,
        roll_degrees: float = 0.0,
    ):
        """
        Mounts an instrument on a platform by setting up the azimuth, elevation and roll of the :ref:`icf`
        with respect to the :ref:`pcf`. Note that we try to use the aircraft definitions of
        azimuth, elevation and roll and we apply the three rotations in the order, azimuth first, elevation second and
        roll third,

        * azimuth is 0.0 when :math:`\\hat{x}_{ICF}` is pointing along the fuselage towards the nose,  i.e. it is parallel to :math:`\\hat{x}_{PCF}`. It is positive towards starboard (right hand).
        * elevation is 0.0 along the fuselage towards the nose when :math:`\\hat{x}_{ICF}` is parallel to :math:`\\hat{x}_{PCF}` and :math:`\\hat{z}_{ICF}` is pointing upwards. It is positive as :math:`\\hat{x}_{ICF}` tilts upwards.
        * roll is 0.0 when :math:`\\hat{y}_{ICF}` is parallel to the wings pointing to port-side (left) and is anti-parallel to :math:`\\hat{y}_{PCF}`. It is positive as :math:`\\hat{y}_{ICF}` tilts upwards and the aircraft leans to the right.

        The instrument control frame is initally configured so the instrument is mounted with azimuth, elevation and roll
        all equal to 0:  :math:`\\hat{x}_{ICF}` is parallel to :math:`\\hat{x}_{PCF}` and :math:`\\hat{z}_{ICF}` is
        anti-parallel to :math:`\\hat{z}_{PCF}`

        Parameters
        ----------
        azimuth_degrees : float
            The azimuth of the instrument boresight in the platform control frame in degrees. Azimuth is zero along the aircraft
            fuselage towards the nose and increases towards the right/starboard
        elevation_degrees : float
            The elevation of the instrument boresight in the platform control frame in degrees. Elevation is
            0.0 along the fuselage towards the nose and increases in the upward direction.
        roll_degrees : float
            The right-handed roll in degrees of the instrument around the final resting point of the boresight, :math:`\\hat{x}_{ICF}`, in the platform control frame. The rotation
            is a right-handed roll around the final. Default is 0.0
        """

        self._ICF_to_PCF.from_yaw_pitch_roll(
            math.radians(azimuth_degrees),
            math.radians(elevation_degrees),
            math.radians(roll_degrees),
        )

    # ------------------------------------------------------------------------------
    #               rotate_instrument_in_icf
    # ------------------------------------------------------------------------------

    def rotate_instrument_in_icf(
        self,
        azimuth_degrees: float,
        elevation_degrees: float,
        roll_degrees: float = 0.0,
    ):
        """
        Sets the instrument rotation matrix so it rotates the instrument control frame, :math:`(\\hat{x}_{ICF}, \\hat{y}_{ICF}, \\hat{z}_{ICF})`,
        through the given azimuth, elevation and roll angles. This method is intended to simulate tilting mirrors and rotation stages
        attached to the instrument. The rotation is applied in the order, azimuth, elevation, roll.

        Parameters
        ----------
        azimuth_degrees : float
            The azimuth of the instrument boresight in the platform control frame in degrees. Azimuth is left handed rotation
            around :math:`\\hat{z}_{ICF}`.
        elevation_degrees : float
            The elevation of the instrument boresight in the platform control frame in degrees. Elevation is
            a left-handed rotation around :math:`\\hat{y}_{ICF}`.
        roll_degrees : float
            The roll in degrees of the instrument. It is a right-handed rotation around :math:`\\hat{x}_{ICF}`. Default is 0.0
        """

        self._IRM.from_azimuth_elevation_roll(
            math.radians(azimuth_degrees),
            math.radians(elevation_degrees),
            math.radians(roll_degrees),
        )

    # -----------------------------------------------------------------------------
    #           orient_platform_in_space
    # -----------------------------------------------------------------------------

    def orient_platform_in_space(
        self, yaw_degrees: float, pitch_degrees: float, roll_degrees: float
    ):
        """
        Sets the platforms yaw, pitch and roll to orient the platform in space at its location. The yaw, pitch
        and roll are applied in a geodetic system so angles are always about the local geodetic system regardless of
        the platform location and the values stay in effect until explicitly changed.

        This function is typically not required in simulators (although it can be set if you wish) as the platform pointing
        in these cases is overridden by the need to look at a "simulated" target location. The function is useful for
        attitude solutions from actual platforms where yaw, pitch and roll are a common method of specifying platform
        orientation

        This function will orient the entire platform in space. The platform control frame is initially set so the
        interpretation of yaw, pitch and roll are sensible, see :ref:`pcf`.

        Parameters
        ----------
        yaw_degrees : float
            The yaw or geographic bearing of the platform control frame in degrees. This is a geographic bearing, N= 0, E= 90, S= 180, W = 270.
        pitch_degrees : float
            The pitch of the platform control frame in degrees. Pitch is right-handed rotation around the starboard pointing :math:`\\hat{y}_{PCF}`.
            A positive pitch will tip the :math:`\\hat{x}_{PCF}` unit vector upwards. Negative values will tip it downwards.
        roll_degrees : float
            The roll of the platform control frame in degrees. Roll is the right-handed rotation around :math:`\\hat{x}_{PCF}`
            after yaw and pitch have been applied. A pilot in a plane looking out of the nose along :math:`\\hat{x}_{PCF}`
            will roll to his right side, (right shoulder falls, left shoulder rises) assuming he is not flying upside down.

        """

        self._PCF_to_GCF.from_yaw_pitch_roll(
            math.radians(yaw_degrees),
            math.radians(pitch_degrees),
            math.radians(roll_degrees),
        )

    # ------------------------------------------------------------------------------
    #           update_location
    # ------------------------------------------------------------------------------
    def set_platform_location(
        self,
        xyzt: tuple[float, float, float, np.datetime64] | None = None,
        latlonheightandt: tuple[float, float, float, np.datetime64] | None = None,
    ):
        """
        Sets the location of the platform to the specified location.

        Parameters
        ----------
        xyzt : Tuple[float, float, float, np.datetime64]
            Set the platform location using geographic geocentric coordinates
        latlonheight : Tuple[float, float, float]
            The the platform location using geodetic corodinates. The tuple is a 4 element array, [0] is latitue, [1] is longitude
            [2] is height in meters, [3] is platform_utc as float representing MJD or numpy.datetime64
        """
        if xyzt is not None:
            self._geo.from_xyz((xyzt[0], xyzt[1], xyzt[2]))
            self._utc = xyzt[3]

        elif latlonheightandt is not None:
            self._geo.from_lat_lon_alt(
                latlonheightandt[0], latlonheightandt[1], latlonheightandt[2]
            )
            self._utc = latlonheightandt[3]
        else:
            logging.warning(
                "set_platform_location, nothing done as neither xyz or latlonheight were set"
            )

        self._geolocation = self._geo.location
        self._local_west = self._geo.local_west
        self._local_south = self._geo.local_south
        self._local_up = self._geo.local_up
        self._GCF_to_ECEF.from_transform_to_destination_coordinates(
            -self._local_west, -self._local_south, self._local_up
        )  #: Transform local geodetic coordinates to geographic geocentric.

    # -----------------------------------------------------------------------------
    #           location
    # -----------------------------------------------------------------------------
    def location(self):
        "returns the geocentric location of the platform. Only valid after a successfull call to set_platform_location"
        return self._geolocation

    # -----------------------------------------------------------------------------
    #           platform_utc
    # -----------------------------------------------------------------------------
    def utc(self):
        return self._utc

    # -----------------------------------------------------------------------------
    #           local_west
    # -----------------------------------------------------------------------------

    def local_west(self):
        """
        Returns the geocentric unit vector of west at the current location of the platform. Only valid after a successful call to set_platform_location
        """
        return self._local_west

    # -----------------------------------------------------------------------------
    #           local_south
    # -----------------------------------------------------------------------------

    def local_south(self):
        """
        Returns the geocentric unit vector of south at the current location of the platform. Only valid after a successful call to set_platform_location
        """
        return self._local_south

    # -----------------------------------------------------------------------------
    #           local_up
    # -----------------------------------------------------------------------------

    def local_up(self):
        """
        Returns the geocentric unit vector of up at the current location of the platform. Only valid after a successful call to set_platform_location
        """
        return self._local_up

    # -----------------------------------------------------------------------------
    #               reset_icf_rotation_matrices
    # -----------------------------------------------------------------------------
    def reset_icf_rotation_matrices(self):
        """
        Resets the IRM and ICF_to_PCF matrices to unity.
        :return:
        """
        self._IRM = RotationMatrix(RotationMatrix.IUnit())
        self._ICF_to_PCF = RotationMatrix(RotationMatrix.IUnit())

    # -----------------------------------------------------------------------------
    #           force_pcf_rotation_matrix
    # -----------------------------------------------------------------------------
    def force_pcf_rotation_matrix(self, GEO: UnitVectors):
        """
        Defines the rotation matrix applied to the platform control frame so the :ref:`icf` unit vectors axes are
        aligned to the ECEF unit-vectors passed in after the full stack of rotation matrices are applied. Note this
        rotation is not applied to the Instrument Rotation Matrix (_IRM) as it is intended to force the primary
        boresight of the instrument control frame point towards a specific location.

        The problem is a linear algebra problem where we have G=RX, where G is the 3x3 Unitvector array passed in,
        X is the initial 3x3 unit vectors in the instrument control frame and R is the stack of rotation matrices. The
        matrix R can be expanded and we get a matrix equation of the form (A )(PCF_to_GCF) ( B )= G where we know
        everything except PCF_to_GCF. Thus we get  PCF_to_GCF = (A-1) G (B-1)


        Parameters
        ----------
        GEO : UnitVectors
            A 3x3 array of column unit vectors. These unit vectors specify the desired orientation of the instrument control frame
            vectors after rotation to the :ref:`ecef`.
        """
        B = RotationMatrix(
            self._PCF_Init_GCF.R  # B is the rotation matrics from instrument to Start of geodetic control frome
            @ self._ICF_to_PCF.R
            @ self._ICF_Init_PCF.R
        )

        A = (
            self._GCF_to_ECEF
        )  # A is the rotation matrix from geodetic control frome to GEO

        self._PCF_to_GCF.from_rotation_matrix(
            A.RInv @ GEO.R @ B.RInv
        )  # get the platform "yaw,pitch roll " rotation matrix

    # -----------------------------------------------------------------------------
    #               convert_icf_to_ecef
    # -----------------------------------------------------------------------------
    def convert_icf_to_ecef(self, v: np.ndarray) -> np.ndarray:
        """
        Converts vectors expressed in the instrument coordinate frame to geographic geocentric ECEF vectors using the
        current rotation matrices.

        Parameters
        ----------
        v : np.ndarray
           An array (3xN) of N vectors expressed in the instrument control frame.

           * v[0,:] is the :math:`\\hat{x}_{ICF}` component of each vector,
           * v[1,:] is the :math:`\\hat{y}_{ICF}` component,
           * v[2,:] is the :math:`\\hat{z}_{ICF}` component.

        Returns
        -------
        np.ndarray
            An array (3xN) of N vectors expressed in the geographic geocentric control frame,

            * v[0,:] is the :math:`\\hat{x}_{ECEF}` component,
            * v[1,:] is the :math:`\\hat{y}_{ECEF}` component,
            * v[2,:] is the :math:`\\hat{z}_{ECEF}` component.
        """
        R = self.get_icf_to_ecef_matrix()
        return R.R @ v

    # ------------------------------------------------------------------------------
    #               convert_icf_to_ecef(self, v: np.ndarray)->np.ndarray:
    # ------------------------------------------------------------------------------

    def convert_icf_to_gcf(self, v: np.ndarray) -> np.ndarray:
        """
        Converts vectors expressed in the instrument coordinate frame to geodetic control frame vectors using the
        current rotation matrices.

        Parameters
        ----------
        v : np.ndarray
            An array (3xN) of N vectors expressed in the instrument control frame.

            * v[0,:] is the :math:`\\hat{x}_{ICF}` component of each vector,
            * v[1,:] is the :math:`\\hat{y}_{ICF}` component,
            * v[2,:] is the :math:`\\hat{z}_{ICF}` component.

        Returns
        -------
        np.ndarray
            An array (3xN) of N vectors expressed in the geodetic control frame,

            * v[0,:] is the :math:`\\hat{x}_{GCF}` (West) component,
            * v[1,:] is the :math:`\\hat{y}_{GCF}` (South) component ,
            * v[2,:] is the :math:`\\hat{z}_{GCF}` (Up) component.
        """

        R = (
            self._PCF_to_GCF.R
            @ self._PCF_Init_GCF.R
            @ self._ICF_to_PCF.R
            @ self._ICF_Init_PCF.R
            @ self._IRM.R
        )
        return R @ v

    # -----------------------------------------------------------------------------
    #           convert_gcf_to_ecef
    # -----------------------------------------------------------------------------

    def convert_gcf_to_ecef(self, v: np.ndarray) -> np.ndarray:
        """
        Converts vectors expressed in the geodetic control frame to geographic geocentric vectors using the
        current rotation matrices.

        Parameters
        ----------
        v : np.ndarray
            An array (3xN) of N vectors expressed in the geodetic control frame.

            * v[0,:] is the :math:`\\hat{x}_{GCF}` component of each vector,
            * v[1,:] is the :math:`\\hat{y}_{GCF}` component,
            * v[2,:] is the :math:`\\hat{z}_{GCF}` component.

        Returns
        -------
        np.ndarray
           An array (3xN) of N vectors expressed in the geodetic control frame,

           * v[0,:] is the :math:`\\hat{x}_{ECEF}` component,
           * v[1,:] is the :math:`\\hat{y}_{ECEF}` component,
           * v[2,:] is the :math:`\\hat{z}_{ECEF}` component.
        """

        return self._GCF_to_ECEF.R @ v

    # -----------------------------------------------------------------------------
    #           get_icf_to_ecef_matrix
    # -----------------------------------------------------------------------------
    def get_icf_to_ecef_matrix(self) -> RotationMatrix:
        """
        Returns the rotation matrix that converts vectors expressed in the instrument control frame to vectors in the geographic geocentric control frame
        If you are inspecting the source code for this method note that the
        rotation matrices are applied in reverse order, ie. right-most or last array in the list is the first rotation applied and left-most or first in the list is the last
        rotation/operation.

        Returns
        -------
        np.ndarray
            An (3x3) rotation matrix. The array is used to transform vectors from the ICF instrument control frame to
            the ECEF geocentric control frame. The matrix should be applied as :math:`\\boldsymbol{V_{GEO}} = \\boldsymbol{R} @ \\boldsymbol{V_{ICF}}`
        """

        full_rotation = (
            self._GCF_to_ECEF.R  # 6) Convert the geodetic vectors to geocentric vectors.
            @ self._PCF_to_GCF.R  # 5) Apply the spatial orientation of the platform in geodetic coordinates
            @ self._PCF_Init_GCF.R  # 4) Rotate the platform control frame so its properly aligned with the geodetic corodinates
            @ self._ICF_to_PCF.R  # 3) Rotate the instrument boresight so its properly positioned on the platform
            @ self._ICF_Init_PCF.R  # 2) Rotate instrument control frame so it is in its initial orientation in the platform control frame
            @ self._IRM.R
        )  # 1) Apply Instrument internal rotations due to mirrors and turntables

        return RotationMatrix(array=full_rotation)



from __future__ import annotations

import logging

import numpy as np

import skretrieval.time as sktime
from skretrieval.core import OpticalGeometry

from .platform_pointing import PlatformPointing
from .rotationmatrix import RotationMatrix


# -----------------------------------------------------------------------------
#           PositionAndOrientationArray
# -----------------------------------------------------------------------------
class PositionAndOrientationArray:
    """
    Stores the array of platform position and orientations used for each instantaneous observation in :class:`~.Platform`.
    This object internally stores a list of platform orientation and states created by calls to

    * :meth:`~.add_current_state` to add a new sample to the policy object
    * :meth:`~.clear` to clear the current list of samples in the policy object

    The instrument model uses the current list of platform states in the platform's :ref:`observationpolicy_class` as one of its inputs
    to the radiative transfer engine. The user must add at least one platform state to the :ref:`observationpolicy_class` for the
    radiative transfer to calculate anything.

    Real observation policies for real instruments may be much more detailed than simply saving platform position,
    orientation and time. For example, mirrors and filters may need to be adjusted, integration times and detector binning
    modes may need to be varied. These aspects are beyond the scope of the platform's observation policy and are left
    for the user to manage and implement.
    """

    def __init__(self):
        """
        Creates a new empty instance.
        """
        self._position: list[
            np.ndarray
        ] = []  # A list [Numsamples] of  array (3) platform_ecef_positions
        self._orientation: list[
            RotationMatrix
        ] = []  # A list [Numsamples] rotation matrices
        self._utc: list[
            float | np.datetime64
        ] = []  # A list [Numsamples] of numpy.datetime64['us']

    # -----------------------------------------------------------------------------
    #           platform_ecef_positions
    # -----------------------------------------------------------------------------

    def ecef_positions(self):
        """
        Returns the geocentric geographic (ECEF) position of the platform for each sample in the observation set.

        Returns
        -------
        np.ndarray( 3,Nt)
            Returns the geographic geocentric position as an array of ECEF vectors, one three element column for each sample in the observation set.
            The position is specified in meters from the centre of the Earth.
        """
        return np.array(self._position).transpose()

    # -----------------------------------------------------------------------------
    #           icf_to_ecef_rotation_matrix
    # -----------------------------------------------------------------------------

    def icf_to_ecef_rotation_matrix(self):
        """
        Returns the orientation of the platform for each sample in the observation set.

        Returns
        -------
        List[RotationMatrix]
            Returns the rotation matrix for each sample. The rotation matric will convert Instrument Control Frame unit
            vectors to geocentric geographic unit vectors. The matrix equation is `vgeo = R.R @ vicf`
        """
        return self._orientation

    # -----------------------------------------------------------------------------
    #           platform_utc
    # -----------------------------------------------------------------------------

    def utc(self):
        """
        Returns the UTC of each of the samples in the observation set. The UTC is either an array of
        numpy.datetime64 values or an array of floats representing MJD.  Note that the datetime64 are detected and converted
        to MJD inside the FrontEndRadianceGenerator before calling Sasktran.
        """
        return np.array(self._utc)

    # -----------------------------------------------------------------------------
    #           numsamples
    # -----------------------------------------------------------------------------

    def numsamples_in_observationset(self) -> int:
        """
        Returns the number of samples in this observation policy

        Returns
        -------
        int
            The number of samples in the observation policy.
        """
        return len(self._utc)

    # -----------------------------------------------------------------------------
    #           clear
    # -----------------------------------------------------------------------------

    def clear(self):
        """
        Clears the current list of platform states inside the observation policy. This method should be called
        before starting a new observation policy
        """
        self._position.clear()
        self._orientation.clear()
        self._utc.clear()

    # -----------------------------------------------------------------------------
    #           add_current_state
    # -----------------------------------------------------------------------------

    def add_current_state(self, platform_attitude: PlatformPointing) -> int:
        """
        Adds the current state of the platform to the observation policy. A typical scenario is for the end-user to
        set the position, orientation and time of the platform using one of many available methods and to then capture the
        platform state for further use within the radiative transfer engine.

        Parameters
        ----------
        platform_attitude: PlatformPointing
            The object used to define the platform position, orientation and time.

        Returns
        -------
        int
            The number of samples in the observation policy.

        """

        R = platform_attitude.get_icf_to_ecef_matrix()
        pos = platform_attitude.location()
        t = platform_attitude.utc()
        self._position.append(pos)
        self._orientation.append(R)
        self._utc.append(t)
        return self.numsamples_in_observationset()

    # ------------------------------------------------------------------------------
    #           to_optical_geometry
    # ------------------------------------------------------------------------------
    def to_optical_geometry(self) -> list[OpticalGeometry]:
        """
        Converts the platform orientations defined in thos PositionAndOrientationArray to a list of OpticalGeometry which can be
        used in retrieval code. This conversion assumes the instrument is looking along the x axis of the :ref:`icf` the instrument
        control frame and the ``local_up`` used by the retrieval is in the direction of the Z axis in the :ref:`icf`.

        Returns
        -------
        List[OpticalGeometry]
            The list of measurement definitions used by retrieval code.
        """

        optical_geometry = []
        for i in range(self.numsamples_in_observationset()):
            obs_pos = self.ecef_positions()[:, i]
            R = self.icf_to_ecef_rotation_matrix()[i].R
            look_vector = R @ np.array([1, 0, 0])
            local_up = R @ np.array([0, 0, 1])
            mjd = sktime.ut_to_mjd(self.utc()[i])
            optical_geometry.append(
                OpticalGeometry(obs_pos, look_vector, local_up, mjd)
            )
        return optical_geometry

    # ------------------------------------------------------------------------------
    #           to_optical_geometry
    # ------------------------------------------------------------------------------
    def to_ecef_position_and_look(
        self, icf_lookvectors=None, one_to_one: bool = False
    ) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
        """
        Converts the platform orientations defined in the PositionAndOrientationArray to a list of OpticalGeometry which can be
        used in retrieval code. This conversion assumes the instrument is looking along the x axis of the :ref:`icf` the instrument
        control frame and the ``local_up`` used by the retrieval is in the direction of the Z axis in the :ref:`icf`.

        Returns
        -------
        Tuple[np.ndarray, np.ndarray, np.ndarray]
            A three element storing (0) the observer ECEF X,Y,Z position, (1) the ECEF look vector away from the instrument/platform  and (2) the UT ttime
        """
        if icf_lookvectors is None:
            icf_lookvectors = np.ndarray([[1.0], [0.0], [0.0]])

        if type(icf_lookvectors) is not np.ndarray:
            icf_lookvectors = np.array(icf_lookvectors)

        s = icf_lookvectors.shape
        numdims = len(s)
        if (numdims == 1) and (s[0] == 3):
            icf_lookvectors = icf_lookvectors.reshape([1, 3])
            s = icf_lookvectors.shape
            numdims = len(s)

        if (numdims != 2) or (s[1] != 3):
            logging.warning(
                "The incoming array of ICF look vectors must be an array of shape (N,3). Your shape is {}".format(  # noqa: G001
                    s
                )
            )

        numinstantaneous_los = s[0]
        num_orientations = self.numsamples_in_observationset()

        if one_to_one:
            ok = numinstantaneous_los == num_orientations
            if not ok:
                msg = "The one_to_one flag can only be used if the number of ICF look vectors {} equals the number of positions and orientations {} ".format(
                    numinstantaneous_los, num_orientations
                )
                raise ValueError(msg)
            numinstantaneous_los = 1

        ecef_observer = np.zeros([num_orientations, numinstantaneous_los, 3])
        ecef_look = np.zeros([num_orientations, numinstantaneous_los, 3])
        ut = np.zeros([num_orientations, numinstantaneous_los], dtype="datetime64[us]")

        for i in range(num_orientations):
            ecef_observer[i, :, :] = self.ecef_positions()[:, i]
            ut[i, :] = sktime.ut_to_datetime64(self.utc()[i])
            R = self.icf_to_ecef_rotation_matrix()[i].R

            if one_to_one:
                look_vector = R @ icf_lookvectors[i, :]
                ecef_look[i, 0, :] = look_vector
            else:
                for j in range(numinstantaneous_los):
                    look_vector = R @ icf_lookvectors[j, :]
                    ecef_look[i, j, :] = look_vector

        if one_to_one:
            ecef_observer = ecef_observer.reshape([num_orientations, 3])
            ecef_look = ecef_look.reshape([num_orientations, 3])

        return (ecef_observer, ecef_look, ut)



from __future__ import annotations

import numpy as np
from numpy.linalg import norm


def gibbs(pos1, pos2, pos3):
    """
    This function uses the Gibbs method of orbit determination from three position vectors. It computes the velocity
    corresponding to the second of three supplied position vectors. It is taken from section 5.2 and Appendix D.10 of
    Orbital Mechanics for Engineering Students, Howard Curtis, 2005, ISBN 0 7506 6169 0.

    Parameters
    ----------
    pos1: np.ndarray(3)
        The ECI position of the first point

    pos2: np.ndarray(3)
        The ECI position of the second point

    pos3: np.ndarray(3)
        The ECI position of the third point.

    Returns
    -------
    np.ndarray(3)
         the velocity corresponding to pos2 in km/s.
    """

    # mu =  398600.0
    mu = 3.986004418e14
    tol = 1e-4

    r1 = norm(pos1)  # Magnitudes of R1, R2 and R3:
    r2 = norm(pos2)
    r3 = norm(pos3)

    c12 = np.cross(pos1, pos2)  # Cross products among pos1, pos2 and pos3:
    c23 = np.cross(pos2, pos3)
    c31 = np.cross(pos3, pos1)

    if (
        np.abs(np.dot(pos1, c23) / r1 / norm(c23)) > tol
    ):  # Check that pos1, pos2 and pos3 are coplanar; if not set error flag:
        V2 = None
    else:
        N = r1 * c23 + r2 * c31 + r3 * c12  # Equation 5.13:
        D = c12 + c23 + c31  # Equation 5.14:
        S = pos1 * (r2 - r3) + pos2 * (r3 - r1) + pos3 * (r1 - r2)  # Equation 5.21:
        V2 = np.sqrt(mu / norm(N) / norm(D)) * (
            np.cross(D, pos2) / r2 + S
        )  # Equation 5.22:

    return V2



from __future__ import annotations

from datetime import datetime, timedelta
from math import pi

import numpy as np
from sgp4.earth_gravity import wgs84
from sgp4.io import twoline2rv

import skretrieval.time

from .satellitebase import SatelliteBase
from .satellitekepler import SatelliteKepler


# -----------------------------------------------------------------------------
#           class SatelliteKepler
# -----------------------------------------------------------------------------
class SatelliteSGP4(SatelliteBase):
    """
    Implements the SGP4 satellite propagator. This is just a shallow wrapper for the `SGP4 python package <https://pypi.org/project/sgp4/>`_
    from PyPi
    """

    def __init__(self, twolines: tuple[str, str] | None = None):
        """
        Creates and initializes the sgp4 orbit predictor using two line elements provided by the user. The lines must be in the
        promper two line format. It is possible to leave the lines as None and initialize the predictor after creation.

        Parameters
        ----------
        twolines: Tuple( str, str )
            Two strings that contain the two line elements used to initialize the sgp4 orbit predictor. This paremeter
            may be None, in which case, the object can be configured by calling :meth:`~.from_kepler_orbit` or :meth:`~.from_twoline_elements`

        """
        super().__init__()
        self.m_sgp4 = None
        if twolines is not None:
            self.from_twoline_elements(twolines[0], twolines[1])

    # -----------------------------------------------------------------------------
    #           from_kepler_orbit
    # -----------------------------------------------------------------------------

    def from_kepler_orbit(self, kepler: SatelliteKepler, bstar=0.0):
        """
        Initializes the sgp4 orbit predictor with two line elements extracted from a kepler orbit.
        A simple algorithm that extracts two line elements from a kepler orbit and feeds them
        into the SGP4 model. Note that the kepler orbit does not correct coordinates to the true equator mean equinox (TEME)
        coordinates used by sgp4 so there will be minor issues for people seeking precision. On the other hand it is a
        simple way to fire up the SGP4 from a Kepler orbit, which is quite useful when modelling simple classes of orbit for
        study work.

        Parameters
        ----------
        kepler : SatelliteKepler
            The Kepler elements for the satellite. The two line elements are generated for the current time and eciposition
            stored in the kep[ler satellite.

        bstar:float
            The B* drag term used by SGP4. This is not part of the Kepler elements and must be manually included by
            the user. We recommend you leave this as default 0.0 unless you know what you are doing as the drag term
            used by SGP4 is tricky to get right (it can change by one or two orders of magnitude for example)

        """
        twolines = kepler.make_two_line_elements(bstar=bstar)
        self.from_twoline_elements(twolines[0], twolines[1])

    # -----------------------------------------------------------------------------
    #           from_twoline_elements
    # -----------------------------------------------------------------------------

    def from_twoline_elements(self, line1: str, line2: str):
        """
        Initializes the sgp4 orbit predictor with classic two line elements.

        Parameters
        ----------
        line1:str
            The first line of the two line elements

        line2:str
            The second line of the two line elements

        """
        self.m_sgp4 = twoline2rv(line1, line2, wgs84)

        orbit_number = int(line2[63 : 63 + 5])  # Get the orbit number of this orbit.
        self.set_orbit_number_from_last_equator_crossing(
            orbit_number, self.m_sgp4.epoch
        )
        self.update_eci_position(self.m_sgp4.epoch)

    # ---------------------------------------------------------------------------
    #             SatelliteKepler::orbital_period
    # ---------------------------------------------------------------------------

    def period(self) -> timedelta:
        """
        Return the orbital period of this orbit using the mean motion

        Returns
        -------
        float
            The orbital period in seconds
        """
        mean_motion = (
            self.m_sgp4.no / 60.0
        )  # Convert sgp4 mean motion from radians/minute to radians per second
        return timedelta(seconds=2.0 * pi / mean_motion)

    # -----------------------------------------------------------------------------
    #           eccentricity
    # -----------------------------------------------------------------------------

    def eccentricity(self) -> float:
        return self.m_sgp4.ecco

    # --------------------------------------------------------------------------
    #            SatelliteKepler::update_eci_position
    # --------------------------------------------------------------------------

    def update_eci_position(self, autc: datetime):
        utc = skretrieval.time.ut_to_datetime(autc)
        if (self._m_time is None) or (
            utc != self._m_time
        ):  # Do we need to update eciposition or is it already the cached value
            r, v = self.m_sgp4.propagate(
                year=utc.year,
                month=utc.month,
                day=utc.day,
                hour=utc.hour,
                minute=utc.minute,
                second=(utc.second + utc.microsecond / 1.0e6),
            )
            pos = np.array(r) * 1000.0
            vel = np.array(v) * 1000.0
            self._set_current_state(pos, vel, utc)



import unittest
import numpy as np
import math
import skretrieval.time as sktime
import sasktran as sk
from skretrieval.platforms.platform import Platform
from skretrieval.platforms.satellite import SatelliteSunSync


class SkretrievalGeometryStatesTests(unittest.TestCase):
    # ------------------------------------------------------------------------------
    #           test_example1
    # ------------------------------------------------------------------------------

    def test_example1(self):
        platform = Platform()
        utc = ["2020-07-15T15:00:00.0000000"]
        observer = [52.0, -107.0, 600000.0]
        tanpoint = [35000.0, 45.0]
        platform.add_measurement_set(
            utc, ("llh", observer), ("tangent_altitude", "limb", tanpoint)
        )
        opticalmeasurements = platform.make_optical_geometry()

        geo = sk.Geodetic()
        for i in range(len(opticalmeasurements)):
            entry = opticalmeasurements[i]
            geo.from_xyz(entry.observer)
            obslat = geo.latitude
            obslng = geo.longitude
            obshgt = geo.altitude / 1000.0
            geo.from_tangent_point(entry.observer, entry.look_vector)
            tanlat = geo.latitude
            tanlng = geo.longitude
            tanhgt = geo.altitude / 1000.0

        self.assertAlmostEqual(obslat, 52.0, 5)
        self.assertAlmostEqual(obslng, 253.0, 5)
        self.assertAlmostEqual(obshgt, 600.0, 5)
        self.assertAlmostEqual(tanlat, 63.64674494609161, 3)
        self.assertAlmostEqual(tanlng, 291.7842681940222, 3)
        self.assertAlmostEqual(tanhgt, 35.000064782094675, 3)

    # ------------------------------------------------------------------------------
    #           test_example2
    # ------------------------------------------------------------------------------

    def test_example2(self):
        platform = Platform()
        utc = ["2020-07-15T15:00:00.0000000"]
        observer = [52.0, -107.0, 35000.0, 45.0, 600000.0]
        tanpoint = [35000.0, 45.0]
        platform.add_measurement_set(
            utc, ("looking_at_llh", observer), ("tangent_altitude", "limb", tanpoint)
        )
        opticalmeasurements = platform.make_optical_geometry()

        geo = sk.Geodetic()
        for i in range(len(opticalmeasurements)):
            entry = opticalmeasurements[i]
            geo.from_xyz(entry.observer)
            obslat = geo.latitude
            obslng = geo.longitude
            obshgt = geo.altitude / 1000.0
            geo.from_tangent_point(entry.observer, entry.look_vector)
            tanlat = geo.latitude
            tanlng = geo.longitude
            tanhgt = geo.altitude / 1000.0
            # print('Satellite location = ({:5.2f}N,{:6.2f}E) at a height of {:6.2f} km'.format(obslat, obslng, obshgt))
            # print('Tangent location = ({:5.2f}N,{:6.2f}E) at a height of {:6.2f} km'.format(tanlat, tanlng, tanhgt))
        self.assertAlmostEqual(obslat, 38.23016297193056, 5)
        self.assertAlmostEqual(obslng, 226.128606362275, 5)
        self.assertAlmostEqual(obshgt, 600.0, 5)
        self.assertAlmostEqual(tanlat, 52.018228238070634, 2)
        self.assertAlmostEqual(tanlng, 252.9849577769396, 2)
        self.assertAlmostEqual(tanhgt, 35.00005703700089, 2)

    # ------------------------------------------------------------------------------
    #           test_position_from_platform
    # ------------------------------------------------------------------------------
    def test_position_from_platform(self):
        mjd0 = sktime.ut_to_mjd(
            "2020-07-15T15:00:00.000000"
        )  # Define the time of the ascending node
        satellite = SatelliteSunSync(
            mjd0,  # Create a sun-synchronous satellite
            orbittype="sgp4",
            period_from_altitude=600000.0,
            localtime_of_ascending_node_hours=2.75,
        )
        platform = Platform(
            platform_locator=satellite
        )  # Create a platform which can use the sun-synchronous satellite for its position

        utc = (
            mjd0 + np.arange(0, 100) / 1440.0
        )  # Get measurements every minute along the orbit starting at the ascending node.
        looktanalt = [
            (35000.0, 45)
        ]  # Look at a tangent altitude of 35 km, at a geographic bearing of 45 degrees at the satellite location. Note, the parameter set will be expanded to N measurements
        platform.add_measurement_set(
            utc, ("from_platform",), ("tangent_altitude", "limb", looktanalt)
        )  # Add the measurement set
        opticalmeasurements = platform.make_optical_geometry()

        geo = sk.Geodetic()
        obslat = np.zeros([100])
        obslng = np.zeros([100])
        obshgt = np.zeros([100])
        tanlat = np.zeros([100])
        tanlng = np.zeros([100])
        tanhgt = np.zeros([100])
        for i in range(len(opticalmeasurements)):
            entry = opticalmeasurements[i]
            geo.from_xyz(entry.observer)
            obslat[i] = geo.latitude
            obslng[i] = geo.longitude
            obshgt[i] = geo.altitude / 1000.0
            geo.from_tangent_point(entry.observer, entry.look_vector)
            tanlat[i] = geo.latitude
            tanlng[i] = geo.longitude
            tanhgt[i] = geo.altitude / 1000.0
            # print('{:3d} Satellite location = ({:5.2f}N,{:6.2f}E) at a height of {:6.2f} km'.format(i, obslat[i], obslng[i], obshgt[i]))
            # print('{:3d} Tangent location   = ({:5.2f}N,{:6.2f}E) at a height of {:6.2f} km'.format(i, tanlat[i], tanlng[i], tanhgt[i]))

        self.assertAlmostEqual(obslat[22], 78.80683736020426, 2)
        self.assertAlmostEqual(obslng[22], 127.3185951847006, 2)
        self.assertAlmostEqual(obshgt[22], 613.2565386016553, 2)
        self.assertAlmostEqual(tanlat[22], 72.68372646991565, 2)
        self.assertAlmostEqual(tanlng[22], 236.5054513231322, 2)
        self.assertAlmostEqual(tanhgt[22], 35.000029306320094, 2)
        self.assertAlmostEqual(obslat[67], -68.11695918550532, 2)
        self.assertAlmostEqual(obslng[67], 319.7019529964772, 2)
        self.assertAlmostEqual(obshgt[67], 625.6432617439019, 2)
        self.assertAlmostEqual(tanlat[67], -48.066050277130465, 2)
        self.assertAlmostEqual(tanlng[67], 344.85005649622474, 2)
        self.assertAlmostEqual(tanhgt[67], 34.99995200953633, 2)

    # ------------------------------------------------------------------------------
    #           test_height_profile
    # ------------------------------------------------------------------------------

    def test_height_profile(self):
        platform = Platform()
        utc = "2020-09-24T18:00:00.0000000"  # Use just a string to define a single UTC
        observer = [
            52.0,
            -107.0,
            10000.0,
            45.0,
            600000.0,
        ]  # define a single observer position
        tanpoint = np.tile(
            [0.0, 45.0], [100, 1]
        )  # define 100 tangent heights, note we use the same geographic bearing as the observer parameters
        tanpoint[:, 0] = np.arange(0, 100) * 1000.0  # fill in the height field
        platform.add_measurement_set(
            utc, ("looking_at_llh", observer), ("tangent_altitude", "limb", tanpoint)
        )  # find the observer position and look vectors etc for the 100 measurements
        opticalmeasurements = platform.make_optical_geometry()

        geo = sk.Geodetic()
        obslat = np.zeros([100])
        obslng = np.zeros([100])
        obshgt = np.zeros([100])
        tanlat = np.zeros([100])
        tanlng = np.zeros([100])
        tanhgt = np.zeros([100])
        for i in range(len(opticalmeasurements)):
            entry = opticalmeasurements[i]
            geo.from_xyz(entry.observer)
            obslat[i] = geo.latitude
            obslng[i] = geo.longitude
            obshgt[i] = geo.altitude / 1000.0
            geo.from_tangent_point(entry.observer, entry.look_vector)
            tanlat[i] = geo.latitude
            tanlng[i] = geo.longitude
            tanhgt[i] = geo.altitude / 1000.0
            # print('self.assertAlmostEqual(obslat[{}], {})'.format(i, obslat[i]))
            # print('self.assertAlmostEqual(obslng[{}], {})'.format(i, obslng[i]))
            # print('self.assertAlmostEqual(obshgt[{}], {})'.format(i, obshgt[i]))
            # print('self.assertAlmostEqual(tanlat[{}], {})'.format(i, tanlat[i]))
            # print('self.assertAlmostEqual(tanlng[{}], {})'.format(i, tanlng[i]))
            # print('self.assertAlmostEqual(tanhgt[{}], {})'.format(i, tanhgt[i]))

        self.assertAlmostEqual(obslat[22], 37.980130977495996, 2)
        self.assertAlmostEqual(obslng[22], 225.5195159438611, 2)
        self.assertAlmostEqual(obshgt[22], 600.0000000000007, 2)
        self.assertAlmostEqual(tanlat[22], 51.91297836680919, 2)
        self.assertAlmostEqual(tanlng[22], 252.62476172260668, 2)
        self.assertAlmostEqual(obslat[67], 37.980130977495996, 2)
        self.assertAlmostEqual(obslng[67], 225.5195159438611, 2)
        self.assertAlmostEqual(obshgt[67], 600.0000000000007, 2)
        self.assertAlmostEqual(tanlat[67], 51.49473602100524, 2)
        self.assertAlmostEqual(tanlng[67], 251.25662163977978, 2)
        self.assertAlmostEqual(tanhgt[67], 67.00005620048381, 2)

    # ------------------------------------------------------------------------------
    #           test_elev_azi
    # ------------------------------------------------------------------------------
    def test_elev_azi(self):
        platform = Platform()
        utc = [
            "2020-09-24T18:00:00.0000000"
        ]  # Use just a string to define a single UTC
        observer = [52.0, -107.0, 10000.0]  # define a single observer position
        yaw_pitch = np.tile([0.0, 0.0], [180, 1])  # define 180 yaw and pitch fields
        yaw_pitch[:, 1] = np.arange(-90, 90)  # fill in the pitch field
        platform.add_measurement_set(
            utc, ("llh", observer), ("yaw_pitch_roll", "standard", yaw_pitch)
        )  # find the observer position and look vectors etc for the 100 measurements
        opticalmeasurements = platform.make_optical_geometry()

        geo = sk.Geodetic()
        obslat = np.zeros([180])
        obslng = np.zeros([180])
        obshgt = np.zeros([180])
        for i in range(len(opticalmeasurements)):
            entry = opticalmeasurements[i]
            geo.from_xyz(entry.observer)
            obslat[i] = geo.latitude
            obslng[i] = geo.longitude
            obshgt[i] = geo.altitude / 1000.0
            up = geo.local_up
            elevation = math.degrees(math.asin(np.dot(up, entry.look_vector)))
            self.assertAlmostEqual(obslat[i], 52.0)
            self.assertAlmostEqual(obslng[i], 360.0 - 107.0)
            self.assertAlmostEqual(obshgt[i], 10.0)
            self.assertAlmostEqual(yaw_pitch[i, 1], elevation)

    # ------------------------------------------------------------------------------
    #           test_instrument_internal_rotation
    # ------------------------------------------------------------------------------
    def test_instrument_internal_rotation(self):
        platform = Platform()
        utc = [
            "2020-09-24T18:00:00.0000000"
        ]  # Use just a string to define a single UTC
        observer = [52.0, -107.0, 10000.0]  # define a single observer position
        azi_elev = np.tile([0.0, 0.0], [180, 1])  # define 180 yaw and pitch fields
        azi_elev[:, 0] = np.arange(0, 180) * 0.1
        platform.add_measurement_set(
            utc,
            ("llh", observer),
            ("azi_elev", "standard", azi_elev),
            icf_orientation=("azi_elev", [30.0, 60.0]),
        )  # find the observer position and look vectors etc for the 100 measurements
        opticalmeasurements = platform.make_optical_geometry()

        geo = sk.Geodetic()
        obslat = np.zeros([180])
        obslng = np.zeros([180])
        obshgt = np.zeros([180])
        for i in range(len(opticalmeasurements)):
            entry = opticalmeasurements[i]
            geo.from_xyz(entry.observer)
            obslat[i] = geo.latitude
            obslng[i] = geo.longitude
            obshgt[i] = geo.altitude / 1000.0
            up = geo.local_up
            north = -geo.local_south
            elevation = math.degrees(math.asin(np.dot(up, entry.look_vector)))
            horiz = entry.look_vector - math.sin(math.radians(elevation)) * up
            horiz /= np.linalg.norm(horiz)
            azi = math.degrees(math.acos(np.dot(north, horiz)))
            self.assertAlmostEqual(obslat[i], 52.0)
            self.assertAlmostEqual(obslng[i], 360.0 - 107.0)
            self.assertAlmostEqual(obshgt[i], 10.0)
            self.assertAlmostEqual(azi, azi_elev[i, 0] + 30.0)
            self.assertAlmostEqual(elevation, 60.0)

    # ------------------------------------------------------------------------------
    #           test_nadir
    # ------------------------------------------------------------------------------
    def test_nadir(self):
        platform = Platform()
        utc = "2020-09-24T18:00:00.0000000"  # Use just a string to define a single UTC
        observer = [52.0, -107.0, 600000.0]  # define a single observer position
        ground = np.tile(
            [57.0, -110.0, 0], [100, 1]
        )  # define 100 tangent heights, note we use the same geographic bearing as the observer parameters
        ground[:, 1] = -107.0 + np.arange(0, 100) / 100.0  # fill in the height field
        platform.add_measurement_set(
            utc, ("llh", observer), ("location_llh", "nadir", ground)
        )  # find the observer position and look vectors etc for the 100 measurements
        opticalmeasurements = platform.make_optical_geometry()

        geo = sk.Geodetic()
        obslat = np.zeros([100])
        obslng = np.zeros([100])
        obshgt = np.zeros([100])
        gndlat = np.zeros([100])
        gndlng = np.zeros([100])
        gndhgt = np.zeros([100])
        for i in range(len(opticalmeasurements)):
            entry = opticalmeasurements[i]
            geo.from_xyz(entry.observer)
            obslat[i] = geo.latitude
            obslng[i] = geo.longitude
            obshgt[i] = geo.altitude / 1000.0
            (l1, l2) = geo.altitude_intercepts(0.0, entry.observer, entry.look_vector)
            geo.from_xyz(l1)
            gndlat[i] = geo.latitude
            gndlng[i] = geo.longitude
            gndhgt[i] = geo.altitude / 1000.0
            self.assertAlmostEqual(obslat[i], 52.0)
            self.assertAlmostEqual(obslng[i], 360.0 - 107.0)
            self.assertAlmostEqual(obshgt[i], 600.0)
            self.assertAlmostEqual(gndlat[i], ground[i, 0], places=4)
            self.assertAlmostEqual(gndlng[i], 360.0 + ground[i, 1], places=4)
            self.assertAlmostEqual(gndhgt[i], ground[i, 2] / 1000.0, places=4)

    # ------------------------------------------------------------------------------
    #           test_tangentalt_geometry
    # ------------------------------------------------------------------------------
    def test_tangent_from_orbitplane_geometry(self):
        mjd0 = sktime.ut_to_mjd("2020-09-24T12:15:36.123456")
        platform = Platform(
            platform_locator=SatelliteSunSync(
                mjd0,
                orbittype="sgp4",
                period_from_altitude=600000.0,
                localtime_of_ascending_node_hours=5.20,
            )
        )
        mjd = mjd0 + np.arange(0, 100) / 1440.0
        looktanalt = [(35000.0, 5.0)]
        platform.add_measurement_set(
            mjd, ("from_platform"), ("tangent_from_orbitplane", "limb", looktanalt)
        )
        opgeom = platform.make_optical_geometry()
        # print("Done")

    # ------------------------------------------------------------------------------
    #           test_tangentalt_geometry
    # ------------------------------------------------------------------------------
    def test_tangentalt_geometry(self):
        platform = Platform()
        utc0 = "2020-09-24T12:15:36.123456"
        platform.platform_locator = SatelliteSunSync(
            utc0,
            orbittype="kepler",
            period_from_altitude=600000.0,
            localtime_of_ascending_node_hours=5.20,
        )

        utc = [utc0, utc0]
        llhobserver = [(52, -107, 600000), (53, -107, 600000)]
        looktanalt = [(35000.0, 45), (35000.0, 45)]
        azielev = [(15, 45.0), (23, 90.0)]

        geodetic = sk.Geodetic()
        geodetic.from_lat_lon_alt(52, -107, 600000.0)
        observer = geodetic.location
        up = geodetic.local_up
        north = -geodetic.local_south
        east = -geodetic.local_west
        plane = north * math.cos(math.radians(45.0)) + east * math.sin(math.radians(45))
        look = geodetic.from_tangent_altitude(35000.0, observer, plane)
        target = geodetic.location

        observer = [observer, observer]
        sixvector = [look[0], look[1], look[2], up[0], up[1], up[2]]
        sixvector = [sixvector, sixvector]
        targetxyz = [target, target]
        lookxyz = [look, look]
        targetllh = [geodetic.latitude, geodetic.longitude, geodetic.altitude]
        targetllh = [targetllh, targetllh]

        platform.add_measurement_set(
            utc, ("llh", llhobserver), ("tangent_altitude", "limb", looktanalt)
        )
        platform.add_measurement_set(
            utc, ("llh", llhobserver), ("tangent_xyz_look", "limb", lookxyz)
        )
        platform.add_measurement_set(
            utc, ("llh", llhobserver), ("location_xyz", "nadir", targetxyz)
        )
        platform.add_measurement_set(
            utc, ("llh", llhobserver), ("location_llh", "nadir", targetllh)
        )
        platform.add_measurement_set(
            utc, ("llh", llhobserver), ("unit_vectors", "none", sixvector)
        )
        platform.add_measurement_set(
            utc, ("llh", llhobserver), ("azi_elev", "standard", azielev)
        )
        platform.add_measurement_set(
            utc, ("llh", llhobserver), ("yaw_pitch_roll", "standard", azielev)
        )
        platform.add_measurement_set(utc, ("llh", llhobserver), ("from_platform"))
        platform.add_measurement_set(
            utc, ("xyz", observer), ("tangent_altitude", "limb", looktanalt)
        )
        platform.add_measurement_set(
            utc, ("from_platform"), ("tangent_altitude", "limb", looktanalt)
        )
        platform.add_measurement_set(
            utc, ("from_platform",), ("tangent_altitude", "limb", looktanalt)
        )
        platform.add_measurement_set(
            utc, ("from_platform", observer), ("tangent_altitude", "limb", looktanalt)
        )
        platform.add_measurement_set(
            utc, ("from_platform", observer), ("from_platform", "limb", looktanalt)
        )
        platform.add_measurement_set(
            utc, ("from_platform", observer), ("from_platform",)
        )
        platform.make_position_and_orientation_array()

    # ------------------------------------------------------------------------------
    #           test_sunsync_satellite
    # ------------------------------------------------------------------------------
    def test_sunsync_satellite(self):
        mjdequator = sktime.ut_to_mjd(
            "2020-07-15 14:50:00.000000"
        )  # Time when satellite crosses equator
        mjd0 = sktime.ut_to_mjd(
            "2020-07-15 15:00:00.000000"
        )  # Time for the start of measurements

        satellite = SatelliteSunSync(
            mjdequator,  # Create the sun-sync satellite
            period_from_altitude=450000.0,
            localtime_of_ascending_node_hours=13.5,
        )

        platform = Platform(
            platform_locator=satellite
        )  # Create the platform using the sun-sync satellite

        twosecs = 2.0 / 86400.0  # Get 2 seconds in MJD days
        tanheights = np.arange(
            8000.0, 40000.0, 250.0
        )  # Get the tangent height of each measurement, 8 km to 40 km in steps of 250 m
        num_meas = tanheights.size

        tanpoint = np.zeros(
            [num_meas, 2]
        )  # specify the parameters for the platform pointing technique
        tanpoint[
            :, 0
        ] = tanheights  # First element is the tangent height the platform will point towards
        tanpoint[
            :, 1
        ] = 0.0  # Second element is the geographic bearing from North. Not quite the same as bearing from orbit plane

        utc = (
            np.arange(num_meas) * twosecs + mjd0
        )  # Get the time of each measurement, 2 seconds apart

        platform.add_measurement_set(
            utc, ("from_platform"), ("tangent_from_orbitplane", "limb", tanpoint)
        )  # Create this measurement set
        platform.make_position_and_orientation_array()
        lookv = platform.icf_to_ecef(np.array([[1], [0], [0]]))
        pos = platform.platform_ecef_positions
        tp = np.zeros([3, num_meas])
        geodetic = sk.Geodetic()
        for i in range(num_meas):
            geodetic.from_tangent_point(pos[:, i], lookv[:, 0, i])
            tp[0, i] = geodetic.latitude
            tp[1, i] = geodetic.longitude
            tp[2, i] = geodetic.altitude
            dh = np.abs(geodetic.altitude - tanheights[i])
            self.assertLess(dh, 0.5)

    # ------------------------------------------------------------------------------
    #           test_backward_looking_system
    # ------------------------------------------------------------------------------
    def test_backward_looking_system(self):
        platform = Platform()
        utc = ["2020-07-15 14:00:00.0000000"]
        observer = [40.0, 0.0, 50000.0, 180.0, 691000.0]
        tanpoint = [50000.0, 45.0]
        platform.add_measurement_set(
            utc, ("looking_at_llh", observer), ("tangent_altitude", "limb", tanpoint)
        )
        obspolicy = platform.make_position_and_orientation_array()
        return platform


if __name__ == "__main__":
    tests = SkretrievalGeometryStatesTests()
    unittest.main()



from __future__ import annotations

import datetime
from collections.abc import Sequence

import astropy.time
import numpy as np
from astropy.utils import iers

from .mjd import utc_to_astropy


# ------------------------------------------------------------------------------
#           gmst
# Calculates the Greenwich Mean Sidereal time.
# ------------------------------------------------------------------------------
def gmst(
    utc: np.datetime64 | (datetime.datetime | (float | np.ndarray)), **kwargs
) -> float | np.ndarray:
    """
    Calculates the Greenwich Mean Sidereal Time for the given Universal time(s). The Universal time is UTC. We use the
    `astropy.time.Time.sidereal_time` function to calculate sidereal time; this function uses IERS corrections for timescales and
    SOFA algorithms. It is is very accurate. The algorithm has given the same results as the novas.sidereal_time function
    to 10 or 11 decimal places.

    Parameters
    ----------
    utc: float, list or array of times.
        A scalar,array or list of coordinated universal times. These Sideral time will be calculated for each of these elements. The
        UTC can be given in any of the formats supported by `juliandate.ut_to_jd` but are typically datetime.datetime
        or numpy.datetime.

    ** kwargs : Optional keyword arguments
        User can pass in optional keyword arguments for method `astropy.time.Time.sideral_time`. Typically used for applications
        that need to specify the IAU algorithm, eg `IAU1982`.

    Returns
    -------
    float or numpy.ndarray
        The greenwich mean sidereal time. This is expressed as a angle in radians between 0 of :math:`2pi`.

    """
    t = utc_to_astropy(utc)
    st = t.sidereal_time("mean", "greenwich", **kwargs)
    return st.radian


# ------------------------------------------------------------------------------
#           rotate_about_zaxis
# ------------------------------------------------------------------------------
def _rotate_about_zaxis(u: np.ndarray, theta: float | np.ndarray) -> np.ndarray:
    v = np.empty_like(u)
    costheta = np.cos(theta)  # do the rotation around the Z axis
    sintheta = np.sin(theta)
    v[0, ...] = costheta * u[0, ...] + sintheta * u[1, ...]
    v[1, ...] = -sintheta * u[0, ...] + costheta * u[1, ...]
    v[2, ...] = u[2, ...]
    return v


# ------------------------------------------------------------------------------
#           polar_motion
# ------------------------------------------------------------------------------
def polar_motion_rotation_matrix(utc, transpose=False) -> tuple[float, float]:
    """
    Returns the polar motion rotation matrix for the given times. The rotation matrix is written as the matrix product
    of :math:`ROT1(y_p)*ROT2(x_p)` where ROT1 is rotation around the global x axis and ROT2 is rotation around the global
    y axis.

    ..  math::

        \\begin{equation}
        \\mathbf{ROT1} =
        \\left( \\begin{array}{ccc}
        \\cos x & 0  & -\\sin x  \\\\
        0      & 1  & 0        \\\\
        \\sin x & 0  & \\cos x
        \\end{array} \\right)
        \\end{equation}

    ..  math::

        \\begin{equation}
        \\mathbf{ROT2} =
        \\left( \\begin{array}{ccc}
        1  & 0       & 0       \\\\
        0  & \\cos y  & \\sin y  \\\\
        0  & -\\sin y & \\cos y
        \\end{array} \\right)
        \\end{equation}

    Parameters
    ----------
    utc : scalar or array
        A scalar or array [N] of coordinated universal times. The number of elements must match the number, `N`, of eci vectors
        The values must be convertible to `astropy.time.Time` values which includes, strings, datetime, numpy.datetime64,
        astropy.time.Time and floats. Float and strings should represent Coordinated Universal Time and floating point values must
        be expressed as a modified julian date.

    transpose : bool
        If true then generate the transpose of the rotation matrix. Typically True is used when rotating
        from Psuedo Earth Fixed to ITRF. False will map ITRF to Psuedo Fixed Earth.

    Returns
    -------
    np.ndarray
        Returns an array of stacked rotation matrices. It will be of shape [3,3] if parameter `utc` is scalar. It will be of shape
        [N,3,3] if parameter `utc` is an array of `N` values.
    """

    t = utc_to_astropy(utc)
    iers_b = iers.IERS_B.open()
    pxarcsec, pyarcsec, status = iers_b.pm_xy(t, return_status=True)

    px = pxarcsec.value * np.pi / 648000.0  # convert arc seconds to radians
    py = pyarcsec.value * np.pi / 648000.0  # convert arc seconds to radians
    cosxp = np.cos(px)
    sinxp = np.sin(px)
    cosyp = np.cos(py)
    sinyp = np.sin(py)

    s = (
        (t.size, 3, 3)
        if (
            (type(utc) is np.ndarray)
            or (type(utc) is astropy.time.Time)
            or (isinstance(utc, Sequence))
        )
        else (3, 3)
    )
    pm = np.zeros(s)
    if transpose:
        pm[..., 0, 0] = cosxp
        pm[..., 0, 1] = sinxp * sinyp
        pm[..., 0, 2] = sinxp * cosyp
        pm[..., 1, 0] = 0
        pm[..., 1, 1] = cosyp
        pm[..., 1, 2] = -sinyp
        pm[..., 2, 0] = -sinxp
        pm[..., 2, 1] = cosxp * sinyp
        pm[..., 2, 2] = cosxp * cosyp
    else:
        pm[..., 0, 0] = cosxp
        pm[..., 0, 1] = 0.0
        pm[..., 0, 2] = -sinxp
        pm[..., 1, 0] = sinxp * sinyp
        pm[..., 1, 1] = cosyp
        pm[..., 1, 2] = cosxp * sinyp
        pm[..., 2, 0] = sinxp * cosyp
        pm[..., 2, 1] = -sinyp
        pm[..., 2, 2] = cosxp * cosyp
    return pm


# ------------------------------------------------------------------------------
#           eci_to_geo
# ------------------------------------------------------------------------------
def eciteme_to_itrf(
    eciv: np.ndarray,
    utc: np.datetime64 | (datetime.datetime | float),
    do_polar_motion: bool = False,
) -> np.ndarray:
    """
    Converts ECI TEME vectors to ITRF/GEO geocentric vectors. Calculates greenwich mean sidereal time using IAU1982 and
    rotates the vector around the Earth's Z axis to get Psuedo Earth Fixed (PEF). This is usually good enough for most
    application and by default does not account for Polar Motion, which is generally quite small, around a few meters.
    Applications that require higher precision can request that polar motion be included but it does slow the code down.

    This code has been tested against the skyfield package using method skyfield.sgp4lib.TEME_to_ITRF.  Agreement is generally
    very good, typically around the centimeter level or better for Low earth satellites. Much of the diffrence seems to be due
    to slight differences in the implementation details of GMST IAU1982. I think the differences are primarily numerical roundoff differences
    rather than algorithm differences. The polar motion has also been checked against skyfield and gives the same answer.


    Parameters
    ----------
    eciv :  np.ndarray
        A 1-D array of size[3,] or a 2-D array of size [3, N] where `N` is the number of vectors. The first column must
        be size 3 and stores the X,Y,Z components of each **ECI** vector.

    utc : scalar or array
        A scalar or array of coordinated universal times. The number of elements must match the number, `N`, of eci vectors
        The values must be convertible to `astropy.time.Time` values which includes, strings, datetime, numpy.datetime64,
        astropy.time.Time and floats. Float and strings should represent Coordinated Universal Time and floating point values must
        be expressed as a modified julian date.

    do_polar_motion : bool
        If true then apply a polar motion correction

    Returns
    -------
    np.ndarray
        Returns an array of geocentric vectors. The array is the same size, shape and units as parameter `eciv`

    References
    ----------
    See `Revisiting Spacetrack Report #3: Rev 1 <https://www.celestrak.com/publications/AIAA/2006-6753/AIAA-2006-6753-Rev1.pdf>`_, D. Vallado,  August 2006, AIAA 2006-6753 Appendix C.
    Also see  function  `TEME_to_ITRF <https://github.com/skyfielders/python-skyfield/blob/master/skyfield/sgp4lib.py>`_
    in package `skyfield.sgp4lib.py` and the matlab function `teme2ecef <https://www.mathworks.com/matlabcentral/fileexchange/62013-sgp4?focused=6160a950-1145-4fa4-b3e7-1f7afc493bee&tab=function>`_
    on Matlab Central. The latter was written in 2007 by David Vallado, who seems to be the civilian expert on all matters relating to SGP4 and TEME ECI.

    """

    if type(eciv) is not np.ndarray:
        eciv = np.array(eciv)
    s = eciv.shape
    assert (
        s[0] == 3
    ), "eciteme_to_itrf, requires the first dimension of the data to be of size 3"

    theta = gmst(
        utc, model="IAU1982"
    )  # Get the Greenwish Mean Sideral time with IAU1982. Return answer in radians.               This will return a scalar or an array N
    geov = _rotate_about_zaxis(
        eciv, theta
    )  # Rotate around the Earths spin axis to get to Psuedo Earth Fixed (PEF), no polar motion.   Returns an array [3,] or an array[3,N]

    if (
        do_polar_motion
    ):  # For highest precision we want to include polar motion. This normally not require
        s = geov.shape  # Get the shape of the vector, save it for later
        if len(s) == 1:  # if we have 1-d column vector {3,}
            geopef = geov  # then use as is.
        else:  # if we have 2 D array [3,N] then
            geopef = (
                geov.transpose()
            )  # transpose it to [N,3] for upcoming matrix multiplication
            geopef = geopef.reshape(
                (s[1], s[0], 1)
            )  # and reshape it to [N,3,1] for upcoming matrix multiplication
        pm = polar_motion_rotation_matrix(
            utc, transpose=True
        )  # Get the polar motion rotation axis, Array[3,3] or [N,3,3]
        geov = (
            pm @ geopef
        )  # do the matrix multiplication (3,3)*(3,)  or stacked multiplication (N,3,3)*(N,3,1) to give (3,) or (N,3)
        geov = np.squeeze(
            geov
        )  # We now have (3,) or (N,3,1) so remove the trailing 1 dimensions (plus any others)
        geov = geov.transpose().reshape(
            s
        )  # transpose back to (3,N) and reshape back to original form and we are done
    return geov  # return the vector


# ------------------------------------------------------------------------------
#           eci_to_geo
# ------------------------------------------------------------------------------
def itrf_to_eciteme(
    geov: np.ndarray,
    utc: np.datetime64 | (datetime.datetime | float),
    do_polar_motion: bool = False,
) -> np.ndarray:
    """
    Converts ITRF/GEO geocentric vectors to ECI TEME vectors. See the sister function :func:`eciteme_to_itrf` for specific details

    Parameters
    ----------
    geov :  np.ndarray
        A 1-D array of size[3,] or a 2-D array of size [3, N] where `N` is the number of vectors. The first column must
        be size 3 and stores the X,Y,Z components of each **ITRF** vector.

    utc : scalar or array
        A scalar or array of coordinated universal times. The number of elements must match the number, `N`, of itrf vectors
        The values must be convertible to `astropy.time.Time` values which includes, strings, datetime, numpy.datetime64,
        astropy.time.Time and floats. Float and strings should represent Coordinated Universal Time and floating point values must
        be expressed as a modified julian date.

    do_polar_motion : bool
        If true then apply a polar motion correction

    Returns
    -------
    np.ndarray
        Returns an array of geocentric vectors. The array is the same size, shape and units as parameter `geov`.

    """
    if type(geov) is not np.ndarray:
        geov = np.array(geov)
    s = geov.shape
    assert (
        s[0] == 3
    ), "itrf_to_eciteme, requires the first dimension of the data to be of size 3"

    if (
        do_polar_motion
    ):  # For highest precision we want to include polar motion. This normally not require
        s = geov.shape  # Get the shape of the vector, save it for later
        if len(s) == 1:  # if we have 1-d column vector {3,}
            geopef = geov  # then use as is.
        else:  # if we have 2 D array [3,N] then
            geopef = (
                geov.transpose()
            )  # transpose it to [N,3] for upcoming matrix multiplication
            geopef = geopef.reshape(
                (s[1], s[0], 1)
            )  # and reshape it to [N,3,1] for upcoming matrix multiplication
        pm = polar_motion_rotation_matrix(
            utc, transpose=False
        )  # Get the polar motion rotation axis, Array[3,3] or [N,3,3]
        geov = (
            pm @ geopef
        )  # do the matrix multiplication (3,3)*(3,)  or stacked multiplication (N,3,3)*(N,3,1) to give (3,) or (N,3)
        geov = np.squeeze(
            geov
        )  # We now have (3,) or (N,3,1) so remove the trailing 1 dimensions (plus any others)
        geov = geov.transpose().reshape(
            s
        )  # transpose back to (3,N) and reshape back to original form and we are done

    theta = -gmst(utc, model="IAU1982")
    return _rotate_about_zaxis(geov, theta)



from __future__ import annotations

import numpy as np
from scipy import sparse
from scipy.optimize import least_squares, minimize
from scipy.optimize._optimize import MemoizeJac

from skretrieval.core.radianceformat import RadianceBase
from skretrieval.retrieval import ForwardModel, Minimizer, RetrievalTarget


class SciPyMinimizer(Minimizer):
    def __init__(
        self,
        method="trf",
        max_nfev=20,
        ftol=1e-3,
        xtol=1e-36,
        x_scale="jac",
        tr_solver="exact",
        include_bounds=False,
        **kwargs,
    ) -> None:
        """
        A minimization wrapper around Scipy's least_squares function

        Parameters
        ----------
        method : str, optional
            Minimization method, see scipy.least_squares, by default "trf"
        max_nfev : int, optional
            Maximum function evalations, see scipy.least_squares, by default 20
        ftol : _type_, optional
            Tolerance on the cost function, see sci, by default 1e-3
        xtol : _type_, optional
            _description_, by default 1e-36
        x_scale : str, optional
            _description_, by default "jac"
        tr_solver : str, optional
            _description_, by default "exact"
        include_bounds : bool, optional
            _description_, by default False
        """
        self._method = method
        self._ftol = ftol
        self._xtol = xtol
        self._max_nfev = max_nfev
        self._x_scale = x_scale
        self._tr_solver = tr_solver
        self._include_bounds = include_bounds

        self._kwargs = kwargs

    def retrieve(
        self,
        measurement_l1: RadianceBase,
        forward_model: ForwardModel,
        retrieval_target: RetrievalTarget,
    ):
        ### Get the prior values
        x_a = retrieval_target.apriori_state()
        initial_guess = retrieval_target.state_vector()
        lb = retrieval_target.lower_bound()
        ub = retrieval_target.upper_bound()

        inv_Sa = retrieval_target.inverse_apriori_covariance()

        if inv_Sa is None:
            # No apriori covariance/regularization
            # Use initial guess to make the matrices as x_a might be None as well
            inv_Sa = np.zeros((len(initial_guess), len(initial_guess)))

        if x_a is None:
            x_a = np.zeros_like(initial_guess)

        ### Get the measurement values
        y_meas_dict = retrieval_target.measurement_vector(measurement_l1)

        y_meas = y_meas_dict["y"]
        good_meas = ~np.isnan(y_meas)
        y_meas = y_meas[good_meas]

        if "y_error" in y_meas_dict:
            # Have measurement error
            if len(np.shape(y_meas_dict["y_error"])) == 1:
                # Only supplied is the diagonal of the error elements
                Sy = sparse.csc_matrix(
                    sparse.diags(y_meas_dict["y_error"][good_meas], 0)
                )
                inv_Sy = sparse.csc_matrix(
                    sparse.diags(1 / y_meas_dict["y_error"][good_meas], 0)
                )
            else:
                Sy = y_meas_dict["y_error"][np.ix_(good_meas, good_meas)]
                inv_Sy = np.linalg.inv(Sy)
        else:
            # No user supplied error, use identity matrix
            Sy = sparse.csc_matrix(sparse.eye(len(y_meas), len(y_meas)))
            inv_Sy = sparse.csc_matrix(sparse.eye(len(y_meas), len(y_meas)))

        y_scaler_inv = sparse.diags(np.sqrt(inv_Sy.diagonal()))

        x_scaler_inv = np.linalg.cholesky(inv_Sa)
        x_scaler = np.linalg.inv(x_scaler_inv)

        y_meas = y_scaler_inv @ y_meas

        x_a = x_scaler_inv @ x_a

        def residual_fun(x):
            retrieval_target.update_state(x_scaler @ x)

            y_ret_dict = retrieval_target.measurement_vector(
                forward_model.calculate_radiance()
            )

            K = y_ret_dict["jacobian"]

            K = y_scaler_inv @ K[good_meas, :] @ x_scaler
            y_ret = y_scaler_inv @ y_ret_dict["y"][good_meas]

            # First part of residuals is from y, y_meas - y_ret, and jacobian K
            res = y_ret - y_meas
            # Second part of residuals is x-x_a, with identity jacobian in scaled space
            res_x = x - x_a
            K_x = np.eye(len(x))

            return np.concatenate((res, res_x)), np.vstack([K, K_x])

        fun = MemoizeJac(residual_fun)
        jac = fun.derivative

        bounds = (
            (lb, ub)
            if self._include_bounds
            else (np.ones_like(lb) * (-np.inf), np.ones_like(ub) * np.inf)
        )

        return least_squares(
            fun,
            x0=x_scaler_inv @ initial_guess,
            jac=jac,
            x_scale=self._x_scale,
            verbose=2,
            tr_solver=self._tr_solver,
            max_nfev=self._max_nfev,
            tr_options={"regularize": False},
            method=self._method,
            xtol=self._xtol,
            ftol=self._ftol,
            bounds=bounds,
            **self._kwargs,
        )


class SciPyMinimizerGrad(Minimizer):
    def __init__(self) -> None:
        super().__init__()

    def retrieve(
        self,
        measurement_l1: RadianceBase,
        forward_model: ForwardModel,
        retrieval_target: RetrievalTarget,
    ):
        ### Get the prior values
        x_a = retrieval_target.apriori_state()
        initial_guess = retrieval_target.state_vector()

        inv_Sa = retrieval_target.inverse_apriori_covariance()

        if inv_Sa is None:
            # No apriori covariance/regularization
            # Use initial guess to make the matrices as x_a might be None as well
            inv_Sa = np.zeros((len(initial_guess), len(initial_guess)))

        if x_a is None:
            x_a = np.zeros_like(initial_guess)

        ### Get the measurement values
        y_meas_dict = retrieval_target.measurement_vector(measurement_l1)

        y_meas = y_meas_dict["y"]
        good_meas = ~np.isnan(y_meas)
        y_meas = y_meas[good_meas]

        if "y_error" in y_meas_dict:
            # Have measurement error
            if len(np.shape(y_meas_dict["y_error"])) == 1:
                # Only supplied is the diagonal of the error elements
                Sy = sparse.csc_matrix(
                    sparse.diags(y_meas_dict["y_error"][good_meas], 0)
                )
                inv_Sy = sparse.csc_matrix(
                    sparse.diags(1 / y_meas_dict["y_error"][good_meas], 0)
                )
            else:
                Sy = y_meas_dict["y_error"][np.ix_(good_meas, good_meas)]
                inv_Sy = np.linalg.inv(Sy)
        else:
            # No user supplied error, use identity matrix
            Sy = sparse.csc_matrix(sparse.eye(len(y_meas), len(y_meas)))
            inv_Sy = sparse.csc_matrix(sparse.eye(len(y_meas), len(y_meas)))

        y_scaler_inv = sparse.diags(np.sqrt(inv_Sy.diagonal()))

        x_scaler_inv = np.linalg.cholesky(inv_Sa)
        x_scaler = np.linalg.inv(x_scaler_inv)

        y_meas = y_scaler_inv @ y_meas

        x_a = x_scaler_inv @ x_a

        def residual_fun(x):
            retrieval_target.update_state(x_scaler @ x)

            y_ret_dict = retrieval_target.measurement_vector(
                forward_model.calculate_radiance()
            )

            K = y_scaler_inv @ y_ret_dict["jacobian"][good_meas, :] @ x_scaler
            y_ret = y_scaler_inv @ y_ret_dict["y"][good_meas]

            cost = (y_ret - y_meas).T @ (y_ret - y_meas).T + (x - x_a).T @ (x - x_a)

            grad = K.T @ (y_ret - y_meas) + (x - x_a)
            K_x = np.eye(len(x))

            full_K = np.vstack([K, K_x])

            return (cost, 2 * grad), 2 * full_K.T @ full_K

        fun = MemoizeJac(residual_fun)
        hess = fun.derivative

        return minimize(
            fun,
            x0=x_scaler_inv @ initial_guess,
            jac=True,
            hess=hess,
            options={"disp": True, "maxiter": 30},
            method="trust-exact",
        )



from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from datetime import datetime

import numpy as np
from astropy import units
from astropy.coordinates import EarthLocation


# ------------------------------------------------------------------------------
#           Class PlatformLocation
# ------------------------------------------------------------------------------
class PlatformLocation(ABC):
    """
    The base class used to specify the location and velocity, if appropriate, of platforms that carry instruments.
    This covers spacecraft, aircraft, balloons and ground sites. The class exposes two abstract functions
    that all classes derived form PlatformLocation must override,

        #. :meth:`~.PlatformLocation.update_position`.
        #. :attr:`~.PlatformLocation.position`

    In addition, platforms that move can override,

        #. :attr:`~.PlatformLocation.velocity`. Optional, defaults to [0.0, 0.0, 0.0]

    This class also provides the platform position as an astropy EarthLocation which allows latitude, longitude and height to be determined.

        #. :attr:`~.PlatformLocation.earth_location`
        #. :attr:`~.PlatformLocation.lat_lon_height`.
    """

    # -----------------------------------------------------------------------------
    #           __init__
    # -----------------------------------------------------------------------------

    def __init__(self):  # noqa: B027
        pass

    # ------------------------------------------------------------------------------
    #           update_position
    # ------------------------------------------------------------------------------

    @abstractmethod
    def update_position(self, utc: datetime | np.datetime64) -> np.ndarray | None:
        """
        Updates the geographic geocentric location of the platform at the given coordinated universal time and returns the
        new position. This is an abstract method and must be implemented in derived child classes.

        Parameters
        ----------
        utc : datetime
            The time at which to update the location of the platform

        Returns
        -------
        np.ndarray, Array[3]
            The three element X,Y,Z geographic geocentric location of the platform
        """
        logging.warning(
            "PlatformLocation.update_position, you are calling the base method which does nothing. You probably want to call the method in a derived class."
        )
        return None
        # raise Exception("Do not call base class method")

    # ------------------------------------------------------------------------------
    #           update_velocity
    # ------------------------------------------------------------------------------

    @abstractmethod
    def update_velocity(self, utc: datetime | np.datetime64) -> np.ndarray | None:
        """
        Updates the geographic geocentric location of the platform at the given coordinated universal time and returns the
        new position. This is an abstract method and must be implemented in derived child classes.

        Parameters
        ----------
        utc : datetime
            The time at which to update the location of the platform

        Returns
        -------
        np.ndarray, Array[3]
            The three element X,Y,Z geographic geocentric location of the platform
        """
        tnow = np.datetime64(utc)
        dt = np.timedelta64(1.0, "s")
        t1 = tnow + dt
        p0 = self.update_position(utc)
        p1 = self.update_position(t1)
        return p1 - p0

    # ------------------------------------------------------------------------------
    #           update_orientation
    # ------------------------------------------------------------------------------

    @abstractmethod
    def update_orientation(
        self, utc: datetime | np.datetime64
    ) -> tuple[np.ndarray, np.ndarray, np.ndarray] | None:
        """
        Updates the orientation of the  at the given time and retursn an x,y,z ITRF unit vector.

        Parameters
        ----------
        utc : datetime
            The time at which to update the location of the platform

        Returns
        -------
        Tuple[ x:np.ndarray(3,),  y:np.ndarray(3,), z:np.ndarray(3,) ]
            A three element storing X,Y,Z unit vectors of the platform orientation. Returns None if there is an issue
            fetching the pointing data at the given time.
        """
        logging.warning(
            "PlatformLocation.update_orientation, you are calling the base method which does nothing. You probably want to call the method in a derived class."
        )
        return None

    # ------------------------------------------------------------------------------
    #           position
    # ------------------------------------------------------------------------------
    @property
    @abstractmethod
    def position(self) -> np.ndarray:
        """
        Returns the geographic, ITRF, geocentric location calculated in the last call to :meth:`~.PlatformLocation.update_position`.
        This is an abstract method and must be implemented in a derived classes.

        Returns
        -------
        np.ndarray(3,)
            A three element array storing the X,Y,Z geocentric location (ITRF ecef, wgs84) of the platform. All numbers are in meters.
        """
        msg = "Do not call base class method"
        raise Exception(msg)

    # -----------------------------------------------------------------------------
    #           earth_location
    # -----------------------------------------------------------------------------
    @property
    def earth_location(self) -> EarthLocation:
        """
        Returns the current position of the platform as an Astropy EarthLocation. This is a convenience function
        wrapped around :attr:`~.position`

        Returns
        -------
        EarthLocation : Location of the platform as an astropy EarthLocation object.

        """
        geo = self.position
        return EarthLocation.from_geocentric(geo[0], geo[1], geo[2], unit=units.meter)

    # ------------------------------------------------------------------------------
    #           lat_lon_height
    # ------------------------------------------------------------------------------
    @property
    def lat_lon_height(self) -> np.ndarray:
        """
        Returns the current platform location as a 3 element array of  geodetic coordinates,
        latitude in degrees, longitude in degrees and height in meters. This is a convenience function
        wrapped around :attr:`~.position`

        Returns
        --------
        numpy.ndarray, Array[3,]
            A 3 element array of [ latitude, longitude, height in meters]

        """
        geoid = self.earth_location
        return np.array([geoid.lat.value, geoid.lon.value, geoid.height.value])

        # ------------------------------------------------------------------------------
        #           velocity
        # ------------------------------------------------------------------------------

    @property
    @abstractmethod
    def velocity(self) -> np.ndarray:
        """
        Returns the geocentric velocity calculated in the last call to :meth:`~.PlatformLocation.update_position`.
        This may be implemented in derived child classes to override the default method which returns an array of zeros,

        Returns
        -------
        np.ndarray, Array[3,]
            The three element X,Y,Z geographic geocentric velocity of the platform. All numbers are in meters/second.
        """
        return np.zeros([3])



from __future__ import annotations

import sasktran as sk
from sasktran.exceptions import wrap_skif_functionfail


class MieAerosolOrbit(sk.OpticalProperty):
    """
    Specialized OpticalProperty which supports Mie Aerosol calculations.

    Parameters
    ----------
    particlesize_climatology : sasktran.Climatology

    species : str
        Molecule to use, one of ['H2SO4', 'ICE', 'WATER']
    """

    def __init__(self, particlesize_climatology, species: str):
        super().__init__("MIEAEROSOL_" + species.upper())

        self._species = species
        self._particlesize_climatology = particlesize_climatology

    @wrap_skif_functionfail
    def _update_opticalproperty(self, **kwargs):
        self._iskopticalproperty.SetProperty(
            "SetParticleSizeClimatology",
            self._particlesize_climatology.skif_object(**kwargs),
        )

    @property
    def particlesize_climatology(self):
        return self._particlesize_climatology

    @particlesize_climatology.setter
    def particlesize_climatology(self, value):
        self._particlesize_climatology = value

    def skif_object(self, **kwargs):
        self._update_opticalproperty(**kwargs)
        return super().skif_object()



from __future__ import annotations

from datetime import datetime, timedelta
from math import (
    acos,
    atan2,
    cos,
    degrees,
    fabs,
    floor,
    log10,
    pi,
    pow,
    radians,
    sin,
    sqrt,
)

import numpy as np

import skretrieval.time as sktime
from skretrieval.time.eci import gmst

from .coe import ClassicalOrbitalElements, coe_from_state_vector
from .gibbs import gibbs
from .satellitebase import SatelliteBase


# -----------------------------------------------------------------------------
#           degrees360
# -----------------------------------------------------------------------------
def degrees360(rad_angle: float) -> float:
    degangle = degrees(rad_angle)
    if degangle < 0.0:
        degangle += 360.0
    return degangle


# -----------------------------------------------------------------------------
#           UnitVector
# -----------------------------------------------------------------------------
def UnitVector(v):
    return v / np.linalg.norm(v)


# -----------------------------------------------------------------------------
#           Magnitude
# -----------------------------------------------------------------------------
def Magnitude(v):
    return np.linalg.norm(v)


# -----------------------------------------------------------------------------
#           power_exponent
# -----------------------------------------------------------------------------
def value_exponent(x):
    if x == 0.0:
        expon = 0
        value = 0.0
        signstr = " "
    else:
        expon = int(floor(log10(fabs(x)) + 1.0))
        value = x / pow(10.0, expon)
        signstr = " " if value >= 0.0 else "-"
    return (value, expon, signstr)


# -----------------------------------------------------------------------------
#           class SatelliteKepler
# -----------------------------------------------------------------------------
class SatelliteKepler(SatelliteBase):
    def __init__(
        self,
        utc: np.datetime64 | (datetime | (float | str)) = None,
        period_from_seconds: float | None = None,
        period_from_altitude: float | None = None,
        period_from_semi_major_axis: float | None = None,
        inclination_radians: float | None = None,
        inclination_is_sun_sync: bool = False,
        mean_anomaly: float = 0.0,
        argument_of_perigee: float = 0.0,
        localtime_of_ascending_node_hours: float | None = None,
        longitude_of_ascending_node_degrees: float | None = None,
        right_ascension_ascending_node: float | None = None,
        eccentricity: float = 0.0000001,
        orbitnumber: int = 0,
    ):
        """
        Implements a simple Kepler orbit around the Earth. Multiple options are provided to describe the required
        orbit. A complete description of the various parameters is given in :meth:`~.from_elements`

        :param utc:  option identical to that described in :meth:`~.from_elements`
        :param period_from_seconds:  keyword option identical to that described in  :meth:`~.from_elements`
        :param period_from_altitude: keyword option identical to that described in  :meth:`~.from_elements`
        :param period_from_semi_major_axis: keyword option identical to that described in  :meth:`~.from_elements`
        :param inclination_radians: keyword option identical to that described in  :meth:`~.from_elements`
        :param inclination_is_sun_sync: keyword option identical to that described in  :meth:`~.from_elements`
        :param mean_anomaly: keyword option identical to that described in  :meth:`~.from_elements`
        :param true_anomaly: keyword option identical to that described in  :meth:`~.from_elements`
        :param argument_of_perigee: keyword option identical to that described in  :meth:`~.from_elements`
        :param localtime_of_ascending_node_hours: keyword option identical to that described in  :meth:`~.from_elements`
        :param longitude_of_ascending_node_degrees: keyword option identical to that described in  :meth:`~.from_elements`
        :param right_ascension_ascending_node: keyword option identical to that described in  :meth:`~.from_elements`
        :param eccentricity: keyword option identical to that described in  :meth:`~.from_elements`
        :param orbitnumber: keyword option identical to that described in  :meth:`~.from_elements`
        """

        super().__init__()

        self.m_epoch: datetime = None  # The time when the kepler elements are defined.
        # self.m_mu: float = 3.98601210E14                        # The gravitational parameter for Earth
        self.m_mu: float = 3.986004418e14  # The gravitational parameter for Earth
        self.m_i: float = None  # inclination in radians
        self.m_raan: float = None  # Right Ascension of ascending node in radians
        self.m_argument_of_perigee: float = None  # Argument of perigee in radians
        self.m_N0: float = 0.0  # Mean motion in radians per second
        self.m_M0: float = (
            0.0  # Mean Anomaly in radians per second at the specified epoch
        )
        self.m_e: float = 0.0  # eccentricity
        self.m_a: float = 0.0  # semi-major axis in meters
        self.m_b: float = 0.0  # semi minor axis in meters
        self.m_h: float = 0.0  # angular momentum of the orbit. This is conserved for a pure kepler orbit
        self.m_xunit: np.ndarray = np.zeros(
            [3]
        )  # Get the unit vector pointing to the perigree (i.e. same direction as eccentricy vector)
        self.m_yunit: np.ndarray = np.zeros(
            [3]
        )  # Get the unit vector pointing along minor axis of ellips, (circa 90 degrees mean anomaly).
        self.m_zunit: np.ndarray = np.zeros(
            [3]
        )  # Get the unit vector perpendicular to the plane (ie in the direction of the angular momentum.

        if (utc is not None) and (
            (period_from_seconds is not None)
            or (period_from_altitude is not None)
            or (period_from_semi_major_axis is not None)
        ):
            utcdatetime = sktime.ut_to_datetime(utc)
            self.from_elements(
                utcdatetime,
                period_from_seconds=period_from_seconds,
                period_from_altitude=period_from_altitude,
                period_from_semi_major_axis=period_from_semi_major_axis,
                inclination_radians=inclination_radians,
                inclination_is_sun_sync=inclination_is_sun_sync,
                mean_anomaly=mean_anomaly,
                argument_of_perigee=argument_of_perigee,
                localtime_of_ascending_node_hours=localtime_of_ascending_node_hours,
                longitude_of_ascending_node_degrees=longitude_of_ascending_node_degrees,
                right_ascension_ascending_node=right_ascension_ascending_node,
                eccentricity=eccentricity,
                orbitnumber=orbitnumber,
            )

    # --------------------------------------------------------------------------
    #        SatelliteKepler::orbital_period
    # --------------------------------------------------------------------------

    def period(self) -> timedelta:
        """
        Return the orbital period of this orbit. Uses keplers third law.

        Returns
        -------
        datetime.timedelta
            The orbital period as a timedelta. Use timedelta method total_seconds() to get the period in seconds.
        """
        return timedelta(
            seconds=2.0 * pi * sqrt(self.m_a * self.m_a * self.m_a / self.m_mu)
        )

    # ------------------------------------------------------------------------------
    #           true_anomaly_to_mean_anomaly
    # ------------------------------------------------------------------------------

    @staticmethod
    def true_anomaly_to_mean_anomaly(nu: float, e: float) -> float:
        """
        Converts True anomaly, :math:`\\nu`, to mean anomaly, :math:`M` by first calculating the eccentric anomaly,
        :math:`E` using the formula,

        ..  math::

            \\tan \\frac{\\nu}{2} = \\sqrt{\\frac{1+e}{1-e}}\\tan\\frac{E}{2}

        and then calculating the mean anomaly with

        ..  math::

            M = E -e\\,\\sin E

        Parameters
        ----------
        nu : float
            True anomaly in radians
        e : float
            eccentricity

        Returns
        -------
        float
            Mean anomaly in radians
        """

        factor = sqrt((1.0 - e) / (1.0 + e))
        y = factor * sin(nu * 0.5)
        x = cos(nu * 0.5)
        E = 2.0 * atan2(y, x)
        return E - e * sin(E)

    # -----------------------------------------------------------------------------
    #           eccentricity
    # -----------------------------------------------------------------------------

    def eccentricity(self) -> float:
        """
        Returns the eccentricity of the orbit ( 0.0 to 1.0)

        """
        return self.m_e

    # --------------------------------------------------------------------------
    #             SatelliteKepler::self.eccentric_anomaly
    # --------------------------------------------------------------------------

    def _eccentric_anomaly(self, M: float, ecc: float, eps: float) -> float:
        """
        Calculate the eccentric anomaly by solving Kepler's equation using a
        Newton-Raphsom method.  Iterate to a precision specified by eps which
        is typically 1.0E-07 to 1.0E-10

        Parameters
        ----------
        M: float
            Mean Anomaly in radians
        ecc: float
            Eccentricity
        eps: float
            Precision, typically 1.0E-07 to 1.0E-10

        Returns
        -------
        float
            The eccentric anomaly in radians
        """

        E = M  # first guess (which is exact for a circle)
        delta = 1.0e20
        while fabs(delta) >= eps:  # and see if we have converged
            delta = E - ecc * sin(E) - M  # estimate the correction
            E -= delta / (1 - ecc * cos(E))  # get the corrected guess
        return E

    # -----------------------------------------------------------------------------
    #           sun_synchronous_inclination
    # -----------------------------------------------------------------------------

    def sun_synchronous_inclination(self, semi_major_axis_meters: float) -> float:
        """
        Method that returns the inclination of a sun synchronous orbit given the semi-major axis.
        For reference see `Sun-Synchronous orbit <https://en.wikipedia.org/wiki/Sun-synchronous_orbit>`_

        Parameters
        ----------
        semi_major_axis_meters: float

        Returns
        -------
        float
            The required inclination of the orbit in radians
        """

        cosi = -pow(
            semi_major_axis_meters / 12352000.0, 3.5
        )  # see https://en.wikipedia.org/wiki/Sun-synchronous_orbit
        if cosi < -1:
            msg = "The semi major axis of the orbit must be less than 12352 km for sun sync to work"
            raise Exception(msg)
        return acos(cosi)

    # --------------------------------------------------------------------------
    #                   SatelliteKepler::update_eci_position
    # --------------------------------------------------------------------------

    def update_eci_position(self, ut: np.datetime64 | datetime):
        """
        Updates the internal ECI position to the given time

        Parameters
        ----------
        ut: datetime, np.datetime64, float, str
            The time at which the new position is required.
        """

        utc = sktime.ut_to_datetime(ut)
        if (self._m_time is None) or (
            utc != self._m_time
        ):  # Do we need to update eciposition or is it already the cached value
            dt = (utc - self.m_epoch).total_seconds()  # Get the difference in seconds
            M = self.m_M0 + self.m_N0 * dt  # Get the new mean Anomaly
            E = self._eccentric_anomaly(
                M, self.m_e, 1.0e-10
            )  # Solve Keplers equation to a precision of 1.0E-10 radians.

            x = self.m_a * (cos(E) - self.m_e)  # Get the x coordinate of the location
            y = self.m_b * sin(E)  # Get the y coordinate of the location
            r = sqrt(x * x + y * y)  # Get the radial distance
            k = self.m_mu / self.m_h
            vx = -k * y / r
            vy = k * (x / r + self.m_e)
            loc = self.m_xunit * x + self.m_yunit * y  # get the eci-location.
            vel = self.m_xunit * vx + self.m_yunit * vy  # get the eci-velocity
            self._set_current_state(loc, vel, utc)  # set the current state

    # -----------------------------------------------------------------------------
    #           make_two_line_elements
    # -----------------------------------------------------------------------------

    def make_two_line_elements(
        self, first_derivative=0.0, second_derivative=0.0, bstar=0.0, satnum=1
    ) -> tuple[str, str]:
        """
        Returns the current orbital elements as two line elements. This is handy if you want to initialize an SGP4
        predictor from a Kepler "starter orbit" using two line elements. The kepler orbit does not directly support
        drag terms, these have to be provided via the `bstar` parameter.

        Parameters
        ----------
        first_derivative: float
            The first derivative term of the SGP4 two line elements. This is usually not used and can be left as the default value of 0.0

        second_derivative: float
            The second derivative term of the SGP4 two line elements. This is usually not used and can be left as the default value of 0.0

        bstar: float
            The drag term used by SGP4. It is usually entered from empirical measurements. A value of 0.0 should disable drag terms in the SGP4 propagator

        satnum: int
            The satellite number. This is provide for convenience

        Returns
        --------
        tuple(str,str)
            Returns the two lines of the elements as a two element tuple of two strings
        """
        t = self.time
        t0 = datetime(t.year, 1, 1)
        dt = t - t0
        year = (t.year) % 100
        dayofyear = dt.total_seconds() / 86400.0 + 1.0  # January 1 is day 1, not day 0
        revsperday = (self.m_N0 * 86400.0) / (2 * pi)
        orbitnum = self.orbit_number(t)[0]
        secondderiv_value, secondderivexponent, secsign = value_exponent(
            second_derivative
        )
        bstar_value, bstarexponent, bstarsign = value_exponent(bstar)

        line1 = "1 {:5d}U 00001A   {:2d}{:12.8f}{:11.8f} {}{:0>5d}{:2d} {}{:0>5d}{:2d} 0    0X".format(
            satnum,
            year,
            dayofyear,
            first_derivative,
            secsign,
            int(fabs(secondderiv_value) * 1.0e5 + 0.5),
            secondderivexponent,
            bstarsign,
            int(fabs(bstar_value) * 1.0e5 + 0.5),
            bstarexponent,
        )
        line2 = "2 {:5d} {:8.4f} {:8.4f} {:07d} {:8.4f} {:8.4f} {:11.8f}{:5d}X".format(
            satnum,
            degrees(self.m_i),
            degrees360(self.m_raan),
            int(self.m_e * 1.0e7 + 0.5),
            degrees360(self.m_argument_of_perigee),
            degrees360(self.m_M0),
            revsperday,
            orbitnum,
        )
        #       Note that the checksum field has not been filled out in either line
        #        print(line1)
        #        print(line2)
        return (line1, line2)

    # --------------------------------------------------------------------------
    #                       SatelliteKepler::from_elements
    # --------------------------------------------------------------------------

    def from_elements(
        self,
        autc: datetime,
        period_from_seconds: float | None = None,
        period_from_altitude: float | None = None,
        period_from_semi_major_axis: float | None = None,
        inclination_radians: float | None = None,
        inclination_is_sun_sync: bool = False,
        mean_anomaly: float = 0.0,
        argument_of_perigee: float = 0.0,
        localtime_of_ascending_node_hours: float | None = None,
        longitude_of_ascending_node_degrees: float | None = None,
        right_ascension_ascending_node: float | None = None,
        eccentricity: float = 0.0000001,  # Eccentricity thats not quite zero avoids divide by zero problems in WXTRACK
        orbitnumber: int = 0,
    ):
        """
        Define the Kepler orbit from the classic 6 `Keplerian elements <https://en.wikipedia.org/wiki/Orbital_elements>`_.

        #. period
        #. inclination
        #. eccentricity
        #. mean anomaly
        #. right ascension of ascending node
        #. argument of perigee

        Several of the orbital elements can be specified in multiple ways. For example the orbital period can be
        defined directly with seconds or it can be defined using the altitude of the satellite. You must use only one
        of the options to specify a parameter otherwise exceptions will/should be raised.

        The caller must ensure that they explictly set the following three orbital elements,

        #. period or altitude must be set.
        #. inclination must be set.
        #. right ascension of the ascending node must be set.

        Leaving the remaining three elements, (i) mean anomaly, (ii) argument of perigee and (iii) eccentricity as default values
        produces a circular orbit with the satellite initially placed at the location of the perigee (which for a
        circular orbit is somewhat undefined and arbitrary!).

        Parameters
        ----------
        platform_utc: datetime.datetime
            The UTC time of the elements

        period_from_seconds : float
            specifies the orbital period in seconds. An alternative to specify the period is with the optional
            parameter `period_from_altitude` or `period_from_semi_major_axis`. One, but only one, of the three methods
            must be used.

        period_from_altitude : float
             specifies the orbital period using the altitude of the satellite in meters. The altitude is nominal as we do not
             account for oblateness of the Earth etc. The radius of the Earth is internally assumed to be 6378000.0 meters.
             An alternative to specify the period is with the optional parameter `period_from_seconds` or `period_from_semi_major_axis`.
             One, but only one, of the three optional methods must be used.

        period_from_semi_major_axis : float
             specifies the orbital period using the semi_major axis of the satellite in meters. An alternative method
             to specify the period is with the optional parameter `period_from_seconds` or `period_from_altitude`. One,
             but only one, of the three methods must be used.

        inclination_radians : float
            Specifes the inclination of the orbit in radians. An alternative method to specify the inclination is with the optional
            parameter `inclination_is_sun_sync`. One, but only one, of the two optional methods must be used.

        inclination_is_sun_sync : bool
             Default False. If True then specify the inclination of the orbit so it is sun-synchronous. This only
             works for orbits below an altitude of ~5974 km,.  An alternative is to specify the inclination with
             the optional parameter `inclination_radians`.  One, but only one, of the two optional methods may be used.
             Note this only sets the inclination of the orbit as a Kepler orbit propagator does not have the oblate Earth
             model (J2) necessary to model sun synchronicity.

        localtime_of_ascending_node_hours: float
            Default: None. Specifies the nominal local time of the ascending node in hours (0-24).  If set then its value
            represents the "hour" at which you want the ascending node for example a floating point value of 18.25 will
            set the node to 18:15 LT. This value is implemented by overwriting the "longitude_of_ascending_node_degrees" keyword
            used to initialize the kepler orbit. It also forces the satellite to be close to the ascending node at time
            `platform_utc` by changing the mean_anomaly value. Default is None.

        longitude_of_ascending_node_degrees: float
            Default None. Specifies the geographic longitude of the ascending node (-180 to 360)at time `platform_utc`. This is an alternative method
            to using parameters `localtime_of_ascending_node_hours` and `right_ascension_ascending_node` to specify the right ascension of the ascending node.
            One, but only one, of the 3 optional methods may be used, if neither option is used the RAAN is set to 0.0.
            If this option is used then the satellite is forced to be close to the ascending node at time `platform_utc` by changing the mean_anomaly value.
            Note that longitude is in degrees not radians.

        right_ascension_ascending_node: float
            Default None. Specifies the Right Ascension of the ascending node in radians (0 to 2.pi). This is an alternative method
            to parameter `longitude_of_ascending_node_degrees` to specify the right ascension of the ascending node.
            One, but only one, of the two optional methods must be used, if neither option is used the RAAN is set to 0.0.

        mean_anomaly: float
            The mean anomaly in radians at time mjd (0-2.pi). This value will be overidden if you use options
            `localtime_of_ascending_node_hours` or `longitude_of_ascending_node_degrees`.

        argument_of_perigee: float
             Default 0. The argument of perigree in radians (0 to 2.pi).

        eccentricity: float
            Default 0. The eccentricity of the orbit (0 to 1)

        orbitnumber: int
            Default 0. The orbit number at the epoch givern by `mjd`

        """
        # ---- determine the period from user supplied period or user supplied altitude

        utc = sktime.ut_to_datetime(autc)
        if period_from_seconds is not None:
            assert (period_from_altitude is None) and (  # noqa: PT018
                period_from_semi_major_axis is None
            ), "You cannot set period from more than one method"
            N0 = 2 * pi / period_from_seconds
        elif period_from_altitude is not None:
            assert (
                period_from_semi_major_axis is None
            ), "You cannot set period from more than one method"
            a = 6378135.0 + period_from_altitude
            N0 = sqrt(self.m_mu / (a * a * a))
        elif period_from_semi_major_axis is not None:
            a = period_from_semi_major_axis
            N0 = sqrt(self.m_mu / (a * a * a))
        else:
            msg = "You must set one method of setting period"
            raise ValueError(msg)
        self.m_N0 = N0  # Get mean motion in radians per second
        self.m_a = pow(self.m_mu / (self.m_N0 * self.m_N0), 1.0 / 3.0)  # Get the semi

        # ---- determine the inclination of the orbit from user supplied inclination or request for sun-synchronous orbit

        if inclination_radians is not None:
            assert (
                inclination_is_sun_sync is False
            ), "You cannot set both inclination_radians and inclination_is_sun_sync"
            I0 = inclination_radians
        elif inclination_is_sun_sync:
            I0 = self.sun_synchronous_inclination(self.m_a)
        else:
            msg = "You must set one of (i) period_from_altitude or (ii) period_from_seconds"
            raise ValueError(msg)

        if localtime_of_ascending_node_hours is not None:
            assert (
                longitude_of_ascending_node_degrees is None
            ), "You cannot set both (i) localtime_of_ascending_node_hours and (ii) longitude_of_ascending_node_degrees "
            ut_ha = (
                float(utc.hour)
                + utc.minute / 60.0
                + (utc.second + utc.microsecond * 1.0e-06) / 3600.0
            )  # hour angle of Greenwich meridian
            deltaha = (
                localtime_of_ascending_node_hours - ut_ha
            )  # get the diffrence between desired hour angle of ascending node and greenwich
            longitude_of_ascending_node_degrees = deltaha * 15.0

        if longitude_of_ascending_node_degrees is not None:
            assert (
                right_ascension_ascending_node is None
            ), "You cannot set both (i) longitude_of_ascending_node_degrees (or localtime_of_ascending_node_hours)  and  (ii) right_ascension_ascending_node"
            st = degrees(gmst(utc))
            ra = (
                st + longitude_of_ascending_node_degrees
            )  # Get RA at the longitude in hours (0-24)
            if ra >= 360.0:
                ra -= 360.0  # put it in the range 0-24
            if ra < 0.0:
                ra += 360.0
            RAAN = radians(ra)  # and then convert to radians
            mean_anomaly = (
                -argument_of_perigee
            )  # Set the mean_anomaly so the satellite is close to the ascending node at the specified time
        elif right_ascension_ascending_node is not None:
            RAAN = right_ascension_ascending_node  # use the provided right ascension
        else:
            RAAN = 0.0  # Use 0 if no option is provided

        self.m_i = I0
        M0 = mean_anomaly
        if M0 < 0.0:
            M0 += 2.0 * pi
        W0 = argument_of_perigee
        E0 = eccentricity

        self.m_argument_of_perigee = argument_of_perigee
        self.m_raan = RAAN
        self.m_M0 = M0
        self.m_e = E0

        eciz = np.array([0.0, 0.0, 1.0])
        ecix = np.array([cos(RAAN), sin(RAAN), 0.0])  # Get vector to ascending node
        ytemp = np.array(
            [cos(RAAN + pi / 2.0), sin(RAAN + pi / 2.0), 0.0]
        )  # Get vector perpendicular to ascending node i the equatorial plane
        eciy = ytemp * cos(I0) + eciz * sin(
            I0
        )  # Get vector perpendicular to ascending node but in the orbital plane
        self.m_zunit = UnitVector(
            np.cross(ecix, eciy)
        )  # Get vector perpendicular to the orbital plane.
        self.m_xunit = UnitVector(
            ecix * cos(W0) + eciy * sin(W0)
        )  # Get the unit vector pointing at the perigee (Which is the semi major axis)
        self.m_yunit = np.cross(self.m_zunit, self.m_xunit)

        e2 = 1.0 - self.m_e * self.m_e
        self.m_b = self.m_a * sqrt(e2)
        self.m_h = sqrt(self.m_mu * self.m_a * e2)

        self._m_time = None
        self.m_epoch = utc
        self.set_orbit_number_from_last_equator_crossing(orbitnumber, self.m_epoch)
        self.update_eci_position(self.m_epoch)

    # ------------------------------------------------------------------------------
    #           from_classical_orbital_elements
    # ------------------------------------------------------------------------------
    def from_classical_orbital_elements(
        self,
        utc: datetime | (np.datetime64 | (float | str)),
        elements: ClassicalOrbitalElements,
        orbitnumber: int = 0,
    ):
        """
        Update this Kepler object so its orbit follows the values given in the Classical Orbital Elements structure.

        Parameters
        ----------
        utc: datetime, np.datetime64, float, str
            The time of the elements

        elements: ClassicalOrbitalElements
            The structure of classical orbital elements

        orbitnumber: int
            Set the orbit number to this value, default 0.
        """
        e = elements.e
        nu = elements.TA
        M = self.true_anomaly_to_mean_anomaly(nu, e)

        self.from_elements(
            utc,
            period_from_semi_major_axis=elements.a,
            inclination_radians=elements.i,
            argument_of_perigee=elements.W,
            right_ascension_ascending_node=elements.RA,
            mean_anomaly=M,
            eccentricity=e,
            orbitnumber=orbitnumber,
        )

    # --------------------------------------------------------------------------
    #                       from_state_vector
    # --------------------------------------------------------------------------

    def from_state_vector(
        self,
        utc: datetime | (np.datetime64 | (float | str)),
        r: np.ndarray,
        v: np.ndarray,
        orbitnumber: int = 0,
    ):
        """
        Update the Kepler object so it follows the orbit defined by the position and velocity state vector.

        Parameters
        ----------
        utc: datetime, np.datetime64, float, str
            The coordinated universal time of the state vector.
        r:np.ndarray [3]
            The ECI position of the satellite in meters
        v:np.ndarray[3]
            The ECI velocity of the satellite in m/s
        orbitnumber: int
            The orbit number at the epoch, default 0
        """

        elements = coe_from_state_vector(r, v)
        self.from_classical_orbital_elements(utc, elements, orbitnumber=orbitnumber)

    # --------------------------------------------------------------------------
    #                       from_state_vector
    # --------------------------------------------------------------------------
    def from_three_positions(
        self,
        utc2: datetime,
        r1: np.ndarray,
        r2: np.ndarray,
        r3: np.ndarray,
        orbitnumber: int = 0,
    ):
        """
        Determine the orbital elements from three position vectors andUpdate the Kepler object so it follows the calculates orbits derived from the specified state vector

        Parameters
        ----------
        utc2: datetime, np.datetime64, float, str
            The coordinated universal time of the second point.
        r1:np.ndarray [3]
            The fiirst ECI position of the satelite in meters
        r2:np.ndarray [3]
            ECI coordinates at epoch (in metres) eciposition of the satellite
        r3:np.ndarray [3]
            ECI coordinates at epoch (in metres) eciposition of the satellite
        orbitnumber: int
            The orbit number at the epoch
        """

        v2 = gibbs(r1, r2, r3)
        self.from_state_vector(utc2, r2, v2, orbitnumber=orbitnumber)



from __future__ import annotations

import unittest

import numpy as np
from sasktran import Geometry, LineOfSight

from skretrieval.core import OpticalGeometry
from skretrieval.core.lineshape import Gaussian
from skretrieval.legacy.core.sensor.spectrograph import Spectrograph


class TestSpectrograph(unittest.TestCase):
    def test_initialization(self):
        wavel_nm = np.linspace(200, 300, 101)

        line_shape = Gaussian(fwhm=1)

        _ = Spectrograph(
            wavel_nm, line_shape, vert_fov=Gaussian(fwhm=1), horiz_fov=Gaussian(fwhm=1)
        )

    def test_line_shape(self):
        wavel_nm = np.linspace(200, 300, 101)

        line_shape = Gaussian(fwhm=1)

        optical_geo = OpticalGeometry(
            observer=np.array([0, 0, 1]),
            look_vector=np.array([0, 1, 0]),
            local_up=np.array([0, 0, 1]),
            mjd=54372,
        )

        spectrograph = Spectrograph(
            wavel_nm, line_shape, Gaussian(fwhm=1), Gaussian(fwhm=1)
        )

        hires_wavel = spectrograph.required_wavelengths(0.01)

        model_geometry = Geometry()

        angles = np.linspace(-0.1, 0.1, 100)

        for a in angles:
            model_geometry.lines_of_sight.append(
                LineOfSight(
                    mjd=54372, observer=[0, 0, 1], look_vector=[0, np.cos(a), np.sin(a)]
                )
            )

        rad = np.zeros((len(hires_wavel), len(angles)))

        rad[300, 50] = 1

        _ = spectrograph.model_radiance(optical_geo, hires_wavel, model_geometry, rad)
